import dill
import os
import sys
from unified_planning.io import PDDLReader

reader = PDDLReader()


def get_feedback(config, instance, seed=0):

    try:
        import up_ac
        path = '/' + os.path.abspath(up_ac.__file__).strip('/__init__.py')
        path += '/utils'
    except ImportError:
        path = os.getcwd().rsplit('up_ac', 1)[0]
        if path[-1] != "/":
            path += "/"
        path += 'up_ac/utils'
    sys.path.append(r"{}".format(path))
    reader._env.error_used_name = False

    fb = \
        dill.load(open(f'{path}/feedback.pkl', 'rb'))

    feedback = fb(config, instance, seed, reader)

    return feedback



"""Irace algorithm configuration interface for unified planning."""
import pandas as pd
import random
import sys
import os
import subprocess
import importlib
import rpy2.robjects as ro
from rpy2.robjects import pandas2ri

from ConfigSpace.hyperparameters import (
    CategoricalHyperparameter,
    UniformFloatHyperparameter,
    UniformIntegerHyperparameter,
)
from ConfigSpace.conditions import (
    InCondition,
    AndConjunction,
    EqualsCondition,
)

sys.path.append(os.path.abspath(os.path.dirname(__file__)))
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from AC_interface import GenericACInterface
from utils.pcs_transform import transform_pcs


class IraceInterface(GenericACInterface):
    """Irace AC interface."""

    def __init__(self):
        """Initialize Irace interface."""
        GenericACInterface.__init__(self)

        if importlib.metadata.version('ConfigSpace') != '0.6.1':
            print('Installing ConfigSpace 0.6.1')
            subprocess.check_call([sys.executable, "-m", "pip", "install",
                                   "ConfigSpace==0.6.1"],
                                  stdout=subprocess.DEVNULL,
                                  stderr=subprocess.DEVNULL)

            os.execv(sys.executable, ['python'] + sys.argv)

    def transform_conf_from_ac(self, engine, configuration,
                               plantype='OneshotPlanner'):
        """
        Transform configuration from algorithm configuration format to engine format.

        This function takes a configuration in the algorithm configuration format, specific to the provided engine,
        and transforms it into the corresponding format for the given engine.

        :param str engine: Name of the engine for which the configuration is being transformed.
        :param dict configuration: A dictionary containing parameter names with their values.

        :return: A dictionary containing the transformed configuration for the specified engine.
        :rtype: dict

        :raises ValueError: If the specified engine is not supported.

        :note:
            The transformation process varies based on the engine type and specific configurations.
        """
        config = transform_pcs(engine, configuration)
        if engine in ('lpg', 'lpg-anytime'):
            del_list = []
            add_list = []
            for pname, pvalue in config.items():
                if pname in self.engine_param_types[engine]:
                    if self.engine_param_types[engine][pname] == 'FLAGS':
                        del_list.append(pname)
                        flag_pname = pname + '=' + config[pname]
                        add_list.append(flag_pname)
                    elif self.engine_param_types[engine][pname] == 'FLAG':
                        if config[pname] == '1':
                            config[pname] = ''
                        else:
                            del_list.append(pname)

            for dl in del_list:
                del config[dl]

            for al in add_list:
                config[al] = ''

        elif engine == 'fast-downward':
            if len(config) == 1:
                pass
            else:
                if config['fast_downward_search_config'] == 'astar':
                     
                    if config['astar_h'] == 'merge_and_shrink':
                        search_option = \
                            config['fast_downward_search_config'] + '('
                        greedy = config['greedy']
                        before_shrinking = config['before_shrinking']
                        before_merging = config['before_merging']
                        if (before_shrinking == 'false'
                                and before_merging == 'false'):
                            before_merging = 'true'
                        max_states = int(float(config['max_states']))
                        threshold_before_merge = \
                            config['threshold_before_merge']

                        search_option += 'merge_and_shrink(merge_strategy='
                        search_option += '''merge_precomputed
                            (merge_tree=linear(variable_order='''
                        search_option += 'reverse_level)),shrink_strategy='
                        search_option += \
                            f'shrink_bisimulation(greedy={greedy}),'
                        search_option += 'label_reduction='
                        search_option += \
                            f'exact(before_shrinking={before_shrinking},'
                        search_option += f'before_merging={before_merging}),'
                        search_option += f'max_states={max_states}'

                        if max_states == 'true': 
                            threshold_before_merge += \
                                ',threshold_before_merge=1))'
                        else:
                            search_option += '))'
                    else:
                        search_option = \
                            config['fast_downward_search_config'] + '('
                        search_option += config['astar_h'] + '())'
                elif (config['fast_downward_search_config']
                        in ('lazy_greedy', 'eager_greedy')):
                    h_def = ''
                    h_name = []
                    search_option = ''
                    skip = 0
                    for i in range(int(config['number_h'])):
                        idx = int(i) + 1
                        h = config[f'h_{idx}']
                        if h in h_name:
                            skip += 1
                            continue
                        h_name.append(h)
                        if idx > len(h_name):
                            idxxx = idx - skip
                        else:
                            idxxx = idx
                        h_def += f'let(h{h_name[idxxx - 1][:2]},{h}(),'
                    search_option += h_def
                    search_option += config['fast_downward_search_config']
                    search_option += '(['
                    for i, hn in enumerate(h_name):
                        search_option += 'h' + hn[:2]
                        if i + 1 < len(h_name):
                            search_option += ','
                        else:
                            search_option += '],'
                    search_option += 'preferred=['
                    prefs = ''
                    used = []
                    for i in range(int(config['nr_preferred'])):
                        idx = int(i) + 1
                        if idx <= int(config['number_h']):
                            h_idx = int(float(config[f'preferred_{idx}'])) - 1
                            if h_idx > len(h_name) - 1:
                                if len(h_name) > 1:
                                    h_idx = random.randint(0, len(h_name) - 1)
                                else:
                                    h_idx = 0
                            if 'h' + h_name[h_idx][:2] not in used not in used:
                                prefs += 'h' + h_name[h_idx][:2]
                        if (idx < len(h_name)
                                and 'h' + h_name[h_idx][:2] not in used
                                and idx != int(config['nr_preferred'])):
                            prefs += ','
                        elif (']' not in prefs
                                and idx == int(config['nr_preferred'])):
                            if ('h' + h_name[h_idx][:2] in used
                                    and prefs[-1] == ','):
                                prefs = prefs[:-1]
                            prefs += '],'
                        used.append('h' + h_name[h_idx][:2])
                    search_option += prefs
                    cost_type = config['cost_type']
                    search_option += f'cost_type={cost_type}))'
                    for _ in range(len(h_name) - 1):
                        search_option += ')'

                elif config['fast_downward_search_config'] == 'lazy_wastar':
                    h_def = ''
                    h_name = []
                    search_option = ''
                    skip = 0
                    for i in range(int(config['number_h'])):
                        idx = int(i) + 1
                        h = config[f'h_{idx}']
                        if h in h_name:
                            skip += 1
                            continue
                        h_name.append(h)
                        if idx > len(h_name):
                            idxxx = idx - skip
                        else:
                            idxxx = idx
                        h_def += f'let(h{h_name[idxxx - 1][:2]},{h}(),'
                    search_option += h_def
                    search_option += config['fast_downward_search_config']
                    search_option += '(['
                    for i, hn in enumerate(h_name):
                        search_option += 'h' + hn[:2]
                        if i + 1 < len(h_name):
                            search_option += ','
                        else:
                            search_option += '],'
                    w = int(float(config['w']))
                    search_option += f'w={w},'
                    search_option += 'preferred=['
                    prefs = ''
                    used = []
                    for i in range(int(config['nr_preferred'])):
                        idx = int(i) + 1    
                        if idx <= int(config['number_h']):
                            h_idx = int(float(config[f'preferred_{idx}'])) - 1
                            if h_idx > len(h_name) - 1:
                                if len(h_name) > 1:
                                    h_idx = random.randint(0, len(h_name) - 1)
                                else:
                                    h_idx = 0
                            if 'h' + h_name[h_idx][:2] not in used not in used:
                                prefs += 'h' + h_name[h_idx][:2]
                        if (idx < len(h_name)
                                and 'h' + h_name[h_idx][:2] not in used
                                and idx != int(config['nr_preferred'])):
                            prefs += ','
                        elif ']' not in prefs and idx == int(config['nr_preferred']):
                            if 'h' + h_name[h_idx][:2] in used and prefs[-1] == ',':
                                prefs = prefs[:-1]
                            prefs += '],'
                        used.append('h' + h_name[h_idx][:2])
                    search_option += prefs
                    cost_type = config['cost_type']
                    search_option += f'cost_type={cost_type}))'
                    for _ in range(len(h_name) - 1):
                        search_option += ')'

                elif config['fast_downward_search_config'] == 'eager':
                    h = config['h_1']
                    h_name = h
                    search_option = ''
                    h_def = f'let(h{h_name[:2]},{h}(),'
                    weight = int(float(config['weight']))
                    search_option += h_def
                    search_option += config['fast_downward_search_config']
                    search_option += '(single(sum([g(),weight(' + f'h{h_name[:2]}'
                    search_option += f',{weight})])),'
                    search_option += f'preferred=[h{h_name[:2]}],'
                    cost_type = config['cost_type']
                    search_option += f'cost_type={cost_type}))'

                if plantype == 'OneshotPlanner':
                    config = {'fast_downward_search_config': search_option}

                elif plantype == 'AnytimePlanner':
                    config = {'fast_downward_anytime_search_config': search_option}

        elif engine == 'symk':

            if len(config) == 1:
                pass
            else:
                if config['fast_downward_search_config'] == 'astar':
                     
                    if config['astar_h'] == 'merge_and_shrink':
                        search_option = config['fast_downward_search_config'] + '('
                        greedy = config['greedy']
                        before_shrinking = config['before_shrinking']
                        before_merging = config['before_merging']
                        if before_shrinking == 'false' and before_merging == 'false':
                            before_merging = 'true'
                        max_states = int(float(config['max_states']))
                        threshold_before_merge = config['threshold_before_merge']

                        search_option += 'merge_and_shrink(merge_strategy='
                        search_option += 'merge_precomputed(merge_tree=linear(variable_order='
                        search_option += 'reverse_level)),shrink_strategy='
                        search_option += f'shrink_bisimulation(greedy={greedy}),'
                        search_option += 'label_reduction='
                        search_option += f'exact(before_shrinking={before_shrinking},'
                        search_option += f'before_merging={before_merging}),'
                        search_option += f'max_states={max_states}'

                        if max_states == 'true': 
                            threshold_before_merge += ',threshold_before_merge=1))'
                        else:
                            search_option += '))'
                    else:
                        search_option = config['fast_downward_search_config'] + '('
                        search_option += config['astar_h'] + '())'

                elif config['fast_downward_search_config'] in ('lazy_greedy', 'eager_greedy'):
                    h_def = ''
                    h_name = []
                    search_option = ''
                    skip = 0
                    for i in range(int(config['number_h'])):
                        idx = int(i) + 1
                        h = config[f'h_{idx}'] + '()'
                        if h in h_name:
                            skip += 1
                            continue
                        h_name.append(h)
                    search_option += config['fast_downward_search_config']
                    search_option += '(['
                    for i, hn in enumerate(h_name):
                        search_option += hn
                        if i + 1 < len(h_name):
                            search_option += ','
                        else:
                            search_option += '],'
                    search_option += 'preferred=['
                    prefs = ''
                    used = []
                    for i in range(int(config['nr_preferred'])):
                        idx = int(i) + 1
                        if idx <= int(config['number_h']):
                            h_idx = int(float(config[f'preferred_{idx}'])) - 1
                            if h_idx > len(h_name) - 1:
                                if len(h_name) > 1:
                                    h_idx = random.randint(0, len(h_name) - 1)
                                else:
                                    h_idx = 0
                            if h_name[h_idx] not in used:
                                prefs += h_name[h_idx]
                        if idx < len(h_name) and h_name[h_idx] not in used and idx != int(config['nr_preferred']):
                            prefs += ','
                        elif ']' not in prefs and idx == int(config['nr_preferred']):
                            if h_name[h_idx] in used and prefs[-1] == ',':
                                prefs = prefs[:-1]
                            prefs += '],'
                        used.append(h_name[h_idx])
                    search_option += prefs
                    cost_type = config['cost_type']
                    search_option += f'cost_type={cost_type})'

                elif config['fast_downward_search_config'] == 'lazy_wastar':
                    h_def = ''
                    h_name = []
                    search_option = ''
                    skip = 0
                    for i in range(int(config['number_h'])):
                        idx = int(i) + 1
                        h = config[f'h_{idx}'] + '()'
                        if h in h_name:
                            skip += 1
                            continue
                        h_name.append(h)
                    search_option += config['fast_downward_search_config']
                    search_option += '(['
                    for i, hn in enumerate(h_name):
                        search_option += hn
                        if i + 1 < len(h_name):
                            search_option += ','
                        else:
                            search_option += '],'
                    w = int(float(config['w']))
                    search_option += f'w={w},'
                    search_option += 'preferred=['
                    prefs = ''
                    used = []
                    for i in range(int(config['nr_preferred'])):
                        idx = int(i) + 1    
                        if idx <= int(config['number_h']):
                            h_idx = int(float(config[f'preferred_{idx}'])) - 1
                            if h_idx > len(h_name) - 1:
                                if len(h_name) > 1:
                                    h_idx = random.randint(0, len(h_name) - 1)
                                else:
                                    h_idx = 0
                            if h_name[h_idx] not in used not in used:
                                prefs += h_name[h_idx]
                        if idx < len(h_name) and h_name[h_idx] not in used and idx != int(config['nr_preferred']):
                            prefs += ','
                        elif ']' not in prefs and idx == int(config['nr_preferred']):
                            if h_name[h_idx] in used and prefs[-1] == ',':
                                prefs = prefs[:-1]
                            prefs += '],'
                        used.append(h_name[h_idx])
                    search_option += prefs
                    cost_type = config['cost_type']
                    search_option += f'cost_type={cost_type})'

                elif config['fast_downward_search_config'] == 'eager':
                    h = config['h_1'] + '()'
                    h_name = h
                    search_option = ''
                    weight = int(float(config['weight']))
                    search_option += config['fast_downward_search_config']
                    search_option += '(single(sum([g(),weight(' + f'{h_name}'
                    search_option += f',{weight})])),'
                    search_option += f'preferred=[{h_name}],'
                    cost_type = config['cost_type']
                    search_option += f'cost_type={cost_type})'

                if plantype == 'OneshotPlanner':
                    config = {'symk_search_config': search_option}

                elif plantype == 'AnytimePlanner':
                    config = {'symk_anytime_search_config': search_option}

        elif engine in ('enhsp', 'enhsp-any'):
            if not isinstance(configuration, str) and 'params' not in configuration:
                config = ''
                for name, value in configuration.items():
                    config += '-' + name[0] + ' ' + value + ' '
                config = {'params': config}
            else:
                config = configuration

        else:
            config = configuration

        return config

    def get_ps_irace(self, param_space):
        """
        Retrieve parameter space information for configuring irace.

        :param param_space: The ConfigSpace object defining the parameter space.
        :type param_space: ConfigSpace.ConfigurationSpace

        :return: A tuple containing:
            - dict: Default values for parameters.
            - bool: Indicates if there are forbidden parameter value combinations.
        :rtype: tuple
        """

        def set_conditional(c, parent, cond_params):
            """
            Set conditions as strings for irace parameter space.

            :param c: The condition to be set.
            :type c: ConfigSpace.conditions
            :param parent: The parent parameter for the condition.
            :type parent: ConfigSpace.Parameter
            :param cond_params: A dictionary to store conditions as strings.
            :type cond_params: dict

            :return: A dictionary with conditions as strings.
            :rtype: dict
            """
            if isinstance(c, InCondition):
                if isinstance(parent, CategoricalHyperparameter):
                    cond_params[c.child.name] = \
                        f' | {parent.name} %in% c({str(c.values)[1:-1]})'
                elif isinstance(parent, UniformFloatHyperparameter):
                    cond_params[c.child.name] = \
                        f' | {parent.name} %in% r({str(c.values)[1:-1]})'
                elif isinstance(parent, UniformIntegerHyperparameter):
                    cond_params[c.child.name] = \
                        f' | {parent.name} %in% i({str(c.values)[1:-1]})'
            elif isinstance(c, EqualsCondition):
                if isinstance(parent, CategoricalHyperparameter):
                    cond_params[c.child.name] = \
                        f' | {parent.name} == "{c.value}"'
                elif isinstance(parent, UniformFloatHyperparameter):
                    cond_params[c.child.name] = \
                        f' | {parent.name} == {c.value}'
                elif isinstance(parent, UniformIntegerHyperparameter):
                    cond_params[c.child.name] = \
                        f' | {parent.name} == {c.value})'
            return cond_params

        params = param_space.get_hyperparameters_dict()
        conditions = param_space.get_conditions()
        cond_params = {}
        for c in conditions:
            parent = c.get_parents()
            parent = parent[0]
            if isinstance(c, AndConjunction):
                cond_params = set_conditional(c, parent, cond_params)
                for cc in c.components[1:]:
                    parent = cc.get_parents()
                    parent = parent[0]
                    and_cond = set_conditional(cc, parent, cond_params)
                    cond_params[cc.child.name] += \
                        ' &&' + and_cond[cc.child.name][2:]
            else:
                cond_params = set_conditional(c, parent, cond_params)

        names = []
        values = []
            
        for _, param in params.items():
            names.append(param.name)
            values.append(param.default_value)

        default_conf = pd.DataFrame([values], columns=names)

        with (ro.default_converter + pandas2ri.converter).context():
            default_conf = ro.conversion.get_conversion().py2rpy(default_conf)

        irace_param_space = ''

        for p, param in params.items():
            if param.name in cond_params:
                condition = cond_params[param.name]
            else:
                condition = ''
            if isinstance(param, CategoricalHyperparameter):
                choices = ''
                for pc in param.choices:
                    choices += f'\"{pc}\", '
                irace_param_space += '\n' + param.name + \
                    ' \"\" c ' + f'({choices[:-2]})' + condition
            elif isinstance(param, UniformFloatHyperparameter):
                irace_param_space += '\n' + param.name + \
                    ' \"\" r ' + f'({param.lower}, {param.upper})' + condition
            elif isinstance(param, UniformIntegerHyperparameter):
                irace_param_space += '\n' + param.name + \
                    ' \"\" i ' + f'({param.lower}, {param.upper})' + condition

        forbidden = ''
        for f in param_space.forbidden_clauses:
            fpair = f.get_descendant_literal_clauses()
            if isinstance(
                fpair[0].hyperparameter, CategoricalHyperparameter) and\
                    isinstance(
                        fpair[1].hyperparameter, CategoricalHyperparameter):
                forbidden += '\n(' + fpair[0].hyperparameter.name + ' == ' + \
                    f'\"{fpair[0].value}\") ' +\
                    '& (' + fpair[1].hyperparameter.name + ' == ' + \
                    f'\"{fpair[1].value}\")'
            elif isinstance(
                    fpair[0].hyperparameter, CategoricalHyperparameter):
                forbidden += '\n(' + fpair[0].hyperparameter.name + ' == ' + \
                    f'\"{fpair[0].value}\") ' +\
                    '& (' + fpair[1].hyperparameter.name + ' == ' + \
                    f'{fpair[1].value})'
            elif isinstance(
                    fpair[1].hyperparameter, CategoricalHyperparameter):
                forbidden += '\n(' + fpair[0].hyperparameter.name + ' == ' + \
                    f'{fpair[0].value}) ' +\
                    '& (' + fpair[1].hyperparameter.name + ' == ' + \
                    f'\"{fpair[1].value}\")'

        forbidden += '\n'

        if len(forbidden) > 2:
            forbiddens = True
        else:
            forbiddens = False

        with open("forbidden.txt", "w") as text_file:
            text_file.write(forbidden)

        self.irace_param_space = irace_param_space

        return default_conf, forbiddens



"""Functions to transform pcs for engines."""


def transform_pcs(engine, configuration):
    """Transform specific engine output to.

    parameter engine: str, name of engine.
    parameter configuration: pcs object, configuration as pcs.
    """
    config = {}

    if engine in ('lpg', 'lpg-anytime'):
        for c in configuration.keys():
            config['-' + c] = str(configuration[c])

    elif engine == 'fast-downward' or \
            engine == 'pyperplan' or \
            engine == 'symk':
        for c in configuration.keys():
            config[c] = str(configuration[c])

    elif engine == 'tamer':
        for c in configuration.keys():
            config[c] = configuration[c]

    elif engine == 'enhsp':
        for c in configuration.keys():
            config[c] = configuration[c]

    elif engine == 'fmap':
        for c in configuration.keys():
            config[c] = configuration[c]

    return config



from functools import wraps
import multiprocessing
import os
import signal
import time


def parametrized(dec):
    def layer(*args, **kwargs):
        def repl(f):
            return dec(f, *args, **kwargs)
        return repl
    return layer


class TimeExceededException(Exception):
    pass


def function_runner(*args, **kwargs):
    """Used as a wrapper function to handle
    returning results on the multiprocessing side"""

    send_end = kwargs.pop("__send_end")
    function = kwargs.pop("__function")
    try:
        result = function(*args, **kwargs)
    except Exception as e:
        send_end.send(e)
        return
    send_end.send(result)


@parametrized
def run_with_timer(func, max_execution_time):
    @wraps(func)
    def wrapper(*args, **kwargs):
        recv_end, send_end = multiprocessing.Pipe(False)
        kwargs["__send_end"] = send_end
        kwargs["__function"] = func
        
        p = multiprocessing.Process(target=function_runner, args=args,
                                    kwargs=kwargs)
        p.start()
        pid = p.pid
        p.join(max_execution_time)
        if p.is_alive():
            p.terminate()
            p.join()
            if p.is_alive():
                os.kill(pid, signal.SIGKILL)
            raise TimeExceededException("Exceeded Execution Time")
        time.sleep(1)
        if p.is_alive():
            p.kill()
            os.kill(pid, signal.SIGKILL)
        result = recv_end.recv()

        if isinstance(result, Exception):
            raise result

        return result

    return wrapper



"""Functions to enforce timeout in cmd call."""

import subprocess
import sys
import tempfile
import os
import signal
import multiprocessing
from functools import wraps
import time
import unified_planning as up
from unified_planning.engines.engine import OperationMode
from unified_planning.engines.results import (
    LogLevel,
    PlanGenerationResult,
    PlanGenerationResultStatus,
)
from unified_planning.io import PDDLWriter
from typing import IO, Callable, Optional, List, Tuple, Union, Any, cast

from types import MethodType
import up_enhsp
import up_fast_downward
import up_symk
import warnings
import pytamer
import up_lpg
from up_tamer.engine import TState
from unified_planning.engines import pddl_planner
import select
import asyncio
from pebble import concurrent


USE_ASYNCIO_ON_UNIX = False
ENV_USE_ASYNCIO = os.environ.get("UP_USE_ASYNCIO_PDDL_PLANNER")
if ENV_USE_ASYNCIO is not None:
    USE_ASYNCIO_ON_UNIX = ENV_USE_ASYNCIO.lower() in ["true", "1"]


def linux_timeout(planner, engine, timelimit, config):
    """Adjust planner cal for AC on Linux.

    parameter engine: str, name of engine.
    parameter timelimit: int, seconds to run for.
    """
    if engine in ('enhsp', 'enhsp-any'):
        path = os.path.abspath(up_enhsp.__file__)
        path = path.rsplit('/', 1)[0]

        # print('config in to', config)
        config = config['params'].split()

        #search = config['search_algorithm']
        #heuristic = config['heuristic']

        if planner.name == 'enhsp':
            def _get_cmd(self, domain_filename: str, problem_filename: str,
                         plan_filename: str):
                base_command = ['ulimit', '-t', f'{timelimit + 5}', ';',
                                'timeout', '-k', '5', '-s', 'SIGKILL',
                                f'{timelimit}s', 'java', '-Xms2g', '-Xmx2g',
                                '-XX:+UseSerialGC', '-jar',
                                f'{path}/ENHSP/enhsp.jar',
                                '-o', domain_filename, '-f', problem_filename,
                                '-sp', plan_filename]

                base_command = self._manage_parameters(base_command)

                base_command = map(str.strip, base_command)
                base_command = ' '.join(base_command)

                #print('base_command', base_command)

                return base_command

            planner._get_cmd = MethodType(_get_cmd, planner)
            planner._solve = MethodType(_solve_enhsp, planner)

        elif planner.name == 'Anytime-enhsp':
            def _get_anytime_cmd(self, domain_filename: str,
                                 problem_filename: str,
                                 plan_filename: str):
                command = ['ulimit', '-t', f'{timelimit + 5}', ';', 'timeout',
                           '-k', '5', '-s', 'SIGKILL',
                           f'{timelimit}s', 'java', '-Xms2g', '-Xmx2g',
                           '-XX:+UseSerialGC', 
                           '-jar', f'{path}/ENHSP/enhsp.jar',
                           '-o', domain_filename, '-f', problem_filename,
                           '-sp', plan_filename, '-s', f'{config[3]}',
                           '-h', f'{config[1]}', '-anytime']

                command = map(str.strip, command)
                command = ' '.join(command)

                return command

            planner._solve = MethodType(_solve_enhsp, planner)
            planner._get_anytime_cmd = MethodType(_get_anytime_cmd, planner)

        elif planner.name == 'SAT-enhsp':
            def _get_cmd(self, domain_filename: str, problem_filename: str,
                         plan_filename: str):
                command = ['ulimit', '-t', f'{timelimit + 5}', ';', 'timeout',
                           '-k', '5', '-s', 'SIGKILL',
                           f'{timelimit}s', 'java', '-Xms2g', '-Xmx2g',
                           '-XX:+UseSerialGC',
                           '-jar', f'{path}/ENHSP/enhsp.jar',
                           '-o', domain_filename, '-f', problem_filename,
                           '-sp', plan_filename, '-s', f'{config[3]}',
                           '-h', f'{config[1]}']

                command = map(str.strip, command)
                command = ' '.join(command)

                return command

            planner._get_cmd = MethodType(_get_cmd, planner)
            planner._solve = MethodType(_solve_enhsp, planner)

        elif planner.name == 'OPT-enhsp':
            def _get_cmd(self, domain_filename: str, problem_filename: str,
                         plan_filename: str):
                command = ['ulimit', '-t', f'{timelimit + 5}', ';', 'timeout',
                           '-k', '5', '-s', 'SIGKILL',
                           f'{timelimit}s', 'java', '-Xms2g', '-Xmx2g',
                           '-XX:+UseSerialGC', 
                           '-jar', f'{path}/ENHSP/enhsp.jar',
                           '-o', domain_filename, '-f', problem_filename,
                           '-sp', plan_filename, '-s', f'{config[3]}',
                           '-h', f'{config[1]}']

                command = map(str.strip, command)
                command = ' '.join(command)

                return command

            planner._get_cmd = MethodType(_get_cmd, planner)
            planner._solve = MethodType(_solve_enhsp, planner)

    elif engine == 'lpg':
        def _get_cmd(self, domain_filename: str, problem_filename: str,
                     plan_filename: str) -> List[str]:
            path = os.path.abspath(up_lpg.__file__)
            path = path.rsplit('/', 1)[0] + '/lpg'
            base_command = 'timeout', '-k', '5', '-s', 'SIGKILL', \
                f'{timelimit}s', path, '-o', domain_filename, '-f', \
                problem_filename, '-n', '1', '-out', plan_filename, \
                *self.parameter + ['-cputime', f'{timelimit}']
            return base_command

        planner._get_cmd = MethodType(_get_cmd, planner)

    elif engine == 'lpg-anytime':
        def _get_anytime_cmd(self, domain_filename: str,
                             problem_filename: str,
                             plan_filename: str) -> List[str]:
            path = os.path.abspath(up_lpg.__file__)
            path = path.rsplit('/', 1)[0] + '/lpg'
            base_command = ['timeout', '-k', '5', '-s', 'SIGKILL', 
                            f'{timelimit}s', path, '-o', domain_filename, '-f',
                            problem_filename, '-n', '1', '-out', plan_filename,
                            *self.parameter] + self._options
            return base_command

        planner._get_anytime_cmd = MethodType(_get_anytime_cmd, planner)

    elif engine == 'fast-downward':

        def _base_cmd(self, plan_filename: str):
            path = os.path.abspath(up_fast_downward.__file__)
            path = path.rsplit('/', 1)[0]
            downward = path + "/downward/fast-downward.py"
            assert sys.executable, "Path to interpreter could not be found"
            cmd = [sys.executable, downward, "--plan-file", plan_filename]
            if self._fd_search_time_limit is not None:
                cmd += ["--search-time-limit", self._fd_search_time_limit]
            # Making sure ff really stops
            cmd += ["--translate-time-limit",
                    f'{timelimit - int(0.1 * timelimit)}']
            cmd += ["--overall-time-limit", f'{timelimit}']
            cmd += ["--log-level", self._log_level]
            return cmd

        planner._base_cmd = MethodType(_base_cmd, planner)

    elif engine == 'symk':

        def _base_cmd(self, plan_filename: str) -> List[str]:
            path = os.path.abspath(up_symk.__file__)
            path = path.rsplit('/', 1)[0]
            downward = path + "/symk/fast-downward.py"
            assert sys.executable, "Path to interpreter could not be found"
            cmd = [sys.executable, downward, "--plan-file", plan_filename]
            if self._symk_search_time_limit is not None:
                cmd += ["--search-time-limit", self._symk_search_time_limit]
            # Making sure ff really stops
            cmd += ["--translate-time-limit",
                    f'{timelimit - int(0.1 * timelimit)}']
            cmd += ["--overall-time-limit", f'{timelimit}']
            cmd += ["--log-level", self._log_level]
            return cmd

        planner._base_cmd = MethodType(_base_cmd, planner)

    return planner


def _solve_enhsp(
    self,
    problem: "up.model.AbstractProblem",
    heuristic: Optional[Callable[["up.model.state.State"],
                        Optional[float]]] = None,
    timeout: Optional[float] = None,
    output_stream: Optional[Union[Tuple[IO[str], IO[str]], IO[str]]] = None,
    anytime=False
) -> "up.engines.results.PlanGenerationResult":
    assert isinstance(problem, up.model.Problem)
    self._writer = PDDLWriter(
        problem, self._needs_requirements, self._rewrite_bool_assignments
    )
    if anytime:
        self._mode_running = OperationMode.ANYTIME_PLANNER
    else:
        self._mode_running = OperationMode.ONESHOT_PLANNER
    plan = None
    logs: List["up.engines.results.LogMessage"] = []
    with tempfile.TemporaryDirectory() as tempdir:
        domain_filename = os.path.join(tempdir, "domain.pddl")
        problem_filename = os.path.join(tempdir, "problem.pddl")
        plan_filename = os.path.join(tempdir, "plan.txt")
        self._writer.write_domain(domain_filename)
        self._writer.write_problem(problem_filename)
        if self._mode_running == OperationMode.ONESHOT_PLANNER:
            cmd = self._get_cmd(domain_filename, problem_filename,
                                plan_filename)
        elif self._mode_running == OperationMode.ANYTIME_PLANNER:
            assert isinstance(
                self, up.engines.pddl_anytime_planner.PDDLAnytimePlanner
            )
            cmd = self._get_anytime_cmd(
                domain_filename, problem_filename, plan_filename
            )
        if output_stream is None:
            # If we do not have an output stream to write to, we simply call
            # a subprocess and retrieve the final output and error with
            # communicate
            process = subprocess.Popen(
                cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
            )
            timeout_occurred: bool = False
            proc_out: List[str] = []
            proc_err: List[str] = []
            try:
                out_err_bytes = process.communicate(timeout=timeout)
                proc_out, proc_err = [[x.decode()] for x in out_err_bytes]
            except subprocess.TimeoutExpired:
                timeout_occurred = True
            retval = process.returncode
        else:
            if sys.platform == "win32":
                # On windows we have to use asyncio (does not work inside notebooks)
                try:
                    loop = asyncio.ProactorEventLoop()
                    exec_res = loop.run_until_complete(
                        pddl_planner.run_command_asyncio(
                            self, cmd, output_stream=output_stream, timeout=timeout
                        )
                    )
                finally:
                    loop.close()
            else:
                # On non-windows OSs, we can choose between asyncio and posix
                # select (see comment on USE_ASYNCIO_ON_UNIX variable for details)
                if USE_ASYNCIO_ON_UNIX:
                    exec_res = asyncio.run(
                        pddl_planner.run_command_asyncio(
                            self, cmd, output_stream=output_stream, timeout=timeout
                        )
                    )
                else:
                    exec_res = run_command_posix_select_enhsp(
                        self, cmd, output_stream=output_stream, timeout=timeout
                    )
            timeout_occurred, (proc_out, proc_err), retval = exec_res

        logs.append(
            up.engines.results.LogMessage(LogLevel.INFO, "".join(proc_out)))
        logs.append(
            up.engines.results.LogMessage(LogLevel.ERROR, "".join(proc_err))
        )
        if os.path.isfile(plan_filename):
            plan = self._plan_from_file(
                problem, plan_filename, self._writer.get_item_named
            )
        if timeout_occurred and retval != 0:
            return PlanGenerationResult(
                PlanGenerationResultStatus.TIMEOUT,
                plan=plan,
                log_messages=logs,
                engine_name=self.name,
            )
    status: PlanGenerationResultStatus = self._result_status(
        problem, plan, retval, logs
    )
    res = PlanGenerationResult(
        status, plan, log_messages=logs, engine_name=self.name
    )
    problem_kind = problem.kind
    if problem_kind.has_continuous_time() or problem_kind.has_discrete_time():
        if isinstance(plan, up.plans.TimeTriggeredPlan) or plan is None:
            return up.engines.results.correct_plan_generation_result(
                res, problem, self._get_engine_epsilon()
            )
    return res


def run_command_posix_select_enhsp(
    engine: pddl_planner.PDDLPlanner,
    cmd: List[str],
    output_stream: Union[Tuple[IO[str], IO[str]], IO[str]],
    timeout: Optional[float] = None,
) -> Tuple[bool, Tuple[List[str], List[str]], int]:
    """
    Executed the specified command line using posix select, imposing the specified timeout and printing online the output on output_stream.
    The function returns a boolean flag telling if a timeout occurred, a pair of string lists containing the captured standard output and standard error and the return code of the command as an integer

    WARNING: this does not work under Windows because the select function only support sockets and not pipes
    WARNING: The resolution of the timeout parameter is ~ 1 second if output_stream is specified
    """
    proc_out: List[str] = []
    proc_err: List[str] = []
    proc_out_buff: List[str] = []
    proc_err_buff: List[str] = []

    #print(cmd)

    engine._process = subprocess.Popen(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True
    )
    
    timeout_occurred: bool = False
    start_time = time.time()
    last_red_out, last_red_err = 0, 0  # Variables needed for the correct loop exit
    readable_streams: List[Any] = []
    # Exit loop condition: Both stream have nothing left to read or the planner is out of time
    while not timeout_occurred and (
        len(readable_streams) != 2 or last_red_out != 0 or last_red_err != 0
    ):
        readable_streams, _, _ = select.select(
            [engine._process.stdout, engine._process.stderr], [], [], 1.0
        )  # 1.0 is the timeout resolution
        if (
            timeout is not None and time.time() - start_time >= timeout
        ):  # Check if the planner is out of time.
            try:
                engine._process.kill()
            except OSError:
                pass  # This can happen if the process is already terminated
            timeout_occurred = True
        for readable_stream in readable_streams:
            out_in_bytes = readable_stream.readline()
            out_str = out_in_bytes.decode().replace("\r\n", "\n")
            if readable_stream == engine._process.stdout:
                if type(output_stream) is tuple:
                    assert len(output_stream) == 2
                    if output_stream[0] is not None:
                        output_stream[0].write(out_str)
                else:
                    cast(IO[str], output_stream).write(out_str)
                last_red_out = len(out_in_bytes)
                buff = proc_out_buff
                lst = proc_out
            else:
                if type(output_stream) is tuple:
                    assert len(output_stream) == 2
                    if output_stream[1] is not None:
                        output_stream[1].write(out_str)
                else:
                    cast(IO[str], output_stream).write(out_str)
                last_red_err = len(out_in_bytes)
                buff = proc_err_buff
                lst = proc_err
            buff.append(out_str)
            if "\n" in out_str:
                lines = "".join(buff).split("\n")
                for x in lines[:-1]:
                    lst.append(x + "\n")

                buff.clear()
                if lines[-1]:
                    buff.append(lines[-1])
        lastout = "".join(proc_out_buff)
        if lastout:
            proc_out.append(lastout + "\n")
        lasterr = "".join(proc_err_buff)
        if lasterr:
            proc_err.append(lasterr + "\n")
    engine._process.wait()
    return timeout_occurred, (proc_out, proc_err), cast(int, engine._process.returncode)



"""Smac algorithm configuration interface for unified planning."""
import random
import sys
import os
import subprocess
import importlib

sys.path.append(os.path.abspath(os.path.dirname(__file__)))
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from AC_interface import GenericACInterface
from utils.pcs_transform import transform_pcs


class SmacInterface(GenericACInterface):
    """Generic Smac interface."""

    def __init__(self):
        """Initialize Smac interface."""
        GenericACInterface.__init__(self)

        if (importlib.metadata.version('ConfigSpace') != '0.6.1' and importlib.
                util.find_spec('selector') is None):
            print('Installing ConfigSpace 0.6.1')
            subprocess.check_call([sys.executable, "-m", "pip", "install",
                                   "ConfigSpace==0.6.1"],
                                  stdout=subprocess.DEVNULL,
                                  stderr=subprocess.DEVNULL)

            os.execv(sys.executable, ['python'] + sys.argv)

    def transform_conf_from_ac(self, engine, configuration, plantype):
        """
        Transform a configuration to the format expected by the planning engines.

        :param str engine: Name of the planning engine.
        :param dict configuration: The configuration with parameter names and values.

        :return: The transformed configuration in the engine's expected format.
        :rtype: dict

        :raises ValueError: If the provided engine list is empty or contains non-string elements.
        """
        if engine in ('lpg', 'lpg-anytime'):
            if isinstance(configuration, dict):
                config = configuration
            else:
                config = configuration.get_dictionary()
            config = transform_pcs(engine, configuration)
            del_list = []
            add_list = []
            for pname, pvalue in config.items():
                if pname in self.engine_param_types[engine]:
                    if self.engine_param_types[engine][pname] == 'FLAGS':
                        del_list.append(pname)
                        flag_pname = pname + '=' + config[pname]
                        add_list.append(flag_pname)
                    elif self.engine_param_types[engine][pname] == 'FLAG':
                        if config[pname] == '1':
                            config[pname] = ''
                        else:
                            del_list.append(pname)

            for dl in del_list:
                del config[dl]

            for al in add_list:
                config[al] = ''

        elif engine == 'fast-downward':
            if isinstance(configuration, dict):
                config = configuration
            else:
                config = configuration.get_dictionary()
            if len(config) == 1:
                pass
            else:
                if config['fast_downward_search_config'] == 'astar':
                     
                    if config['astar_h'] == 'merge_and_shrink':
                        search_option = config['fast_downward_search_config'] + '('
                        greedy = config['greedy']
                        before_shrinking = config['before_shrinking']
                        before_merging = config['before_merging']
                        if before_shrinking == 'false' and before_merging == 'false':
                            before_merging = 'true'
                        max_states = int(float(config['max_states']))
                        threshold_before_merge = config['threshold_before_merge']

                        search_option += 'merge_and_shrink(merge_strategy='
                        search_option += 'merge_precomputed(merge_tree=linear(variable_order='
                        search_option += 'reverse_level)),shrink_strategy='
                        search_option += f'shrink_bisimulation(greedy={greedy}),'
                        search_option += 'label_reduction='
                        search_option += f'exact(before_shrinking={before_shrinking},'
                        search_option += f'before_merging={before_merging}),'
                        search_option += f'max_states={max_states}'

                        if max_states == 'true': 
                            threshold_before_merge += ',threshold_before_merge=1))'
                        else:
                            search_option += '))'
                    else:
                        search_option = config['fast_downward_search_config'] + '('
                        search_option += config['astar_h'] + '())'
                elif config['fast_downward_search_config'] in ('lazy_greedy', 'eager_greedy'):
                    h_def = ''
                    h_name = []
                    search_option = ''
                    skip = 0
                    for i in range(int(config['number_h'])):
                        idx = int(i) + 1
                        h = config[f'h_{idx}']
                        if h in h_name:
                            skip += 1
                            continue
                        h_name.append(h)
                        if idx > len(h_name):
                            idxxx = idx - skip
                        else:
                            idxxx = idx
                        h_def += f'let(h{h_name[idxxx - 1][:2]},{h}(),'
                    search_option += h_def
                    search_option += config['fast_downward_search_config']
                    search_option += '(['
                    for i, hn in enumerate(h_name):
                        search_option += 'h' + hn[:2]
                        if i + 1 < len(h_name):
                            search_option += ','
                        else:
                            search_option += '],'
                    search_option += 'preferred=['
                    prefs = ''
                    used = []
                    for i in range(int(config['nr_preferred'])):
                        idx = int(i) + 1
                        if idx <= int(config['number_h']):
                            h_idx = int(float(config[f'preferred_{idx}'])) - 1
                            if h_idx > len(h_name) - 1:
                                if len(h_name) > 1:
                                    h_idx = random.randint(0, len(h_name) - 1)
                                else:
                                    h_idx = 0
                            if 'h' + h_name[h_idx][:2] not in used not in used:
                                prefs += 'h' + h_name[h_idx][:2]
                        if idx < len(h_name) and 'h' + h_name[h_idx][:2] not in used and idx != int(config['nr_preferred']):
                            prefs += ','
                        elif ']' not in prefs and idx == int(config['nr_preferred']):
                            if 'h' + h_name[h_idx][:2] in used and prefs[-1] == ',':
                                prefs = prefs[:-1]
                            prefs += '],'
                        used.append('h' + h_name[h_idx][:2])
                    search_option += prefs
                    cost_type = config['cost_type']
                    search_option += f'cost_type={cost_type}))'
                    for _ in range(len(h_name) - 1):
                        search_option += ')'

                elif config['fast_downward_search_config'] == 'lazy_wastar':
                    h_def = ''
                    h_name = []
                    search_option = ''
                    skip = 0
                    for i in range(int(config['number_h'])):
                        idx = int(i) + 1
                        h = config[f'h_{idx}']
                        if h in h_name:
                            skip += 1
                            continue
                        h_name.append(h)
                        if idx > len(h_name):
                            idxxx = idx - skip
                        else:
                            idxxx = idx
                        h_def += f'let(h{h_name[idxxx - 1][:2]},{h}(),'
                    search_option += h_def
                    search_option += config['fast_downward_search_config']
                    search_option += '(['
                    for i, hn in enumerate(h_name):
                        search_option += 'h' + hn[:2]
                        if i + 1 < len(h_name):
                            search_option += ','
                        else:
                            search_option += '],'
                    w = int(float(config['w']))
                    search_option += f'w={w},'
                    search_option += 'preferred=['
                    prefs = ''
                    used = []
                    for i in range(int(config['nr_preferred'])):
                        idx = int(i) + 1    
                        if idx <= int(config['number_h']):
                            h_idx = int(float(config[f'preferred_{idx}'])) - 1
                            if h_idx > len(h_name) - 1:
                                if len(h_name) > 1:
                                    h_idx = random.randint(0, len(h_name) - 1)
                                else:
                                    h_idx = 0
                            if 'h' + h_name[h_idx][:2] not in used not in used:
                                prefs += 'h' + h_name[h_idx][:2]
                        if idx < len(h_name) and 'h' + h_name[h_idx][:2] not in used and idx != int(config['nr_preferred']):
                            prefs += ','
                        elif ']' not in prefs and idx == int(config['nr_preferred']):
                            if 'h' + h_name[h_idx][:2] in used and prefs[-1] == ',':
                                prefs = prefs[:-1]
                            prefs += '],'
                        used.append('h' + h_name[h_idx][:2])
                    search_option += prefs
                    cost_type = config['cost_type']
                    search_option += f'cost_type={cost_type}))'
                    for _ in range(len(h_name) - 1):
                        search_option += ')'

                elif config['fast_downward_search_config'] == 'eager':
                    h = config['h_1']
                    h_name = h
                    search_option = ''
                    h_def = f'let(h{h_name[:2]},{h}(),'
                    weight = int(float(config['weight']))
                    search_option += h_def
                    search_option += config['fast_downward_search_config']
                    search_option += '(single(sum([g(),weight(' + f'h{h_name[:2]}'
                    search_option += f',{weight})])),'
                    search_option += f'preferred=[h{h_name[:2]}],'
                    cost_type = config['cost_type']
                    search_option += f'cost_type={cost_type}))'

                if plantype == 'OneshotPlanner':
                    config = {'fast_downward_search_config': search_option}

                elif plantype == 'AnytimePlanner':
                    config = {'fast_downward_anytime_search_config': search_option}

        elif engine == 'symk':

            if isinstance(configuration, dict):
                config = configuration
            else:
                config = configuration.get_dictionary()

            if len(config) == 1:
                pass
            else:
                if config['fast_downward_search_config'] == 'astar':
                     
                    if config['astar_h'] == 'merge_and_shrink':
                        search_option = config['fast_downward_search_config'] + '('
                        greedy = config['greedy']
                        before_shrinking = config['before_shrinking']
                        before_merging = config['before_merging']
                        if before_shrinking == 'false' and before_merging == 'false':
                            before_merging = 'true'
                        max_states = int(float(config['max_states']))
                        threshold_before_merge = config['threshold_before_merge']

                        search_option += 'merge_and_shrink(merge_strategy='
                        search_option += 'merge_precomputed(merge_tree=linear(variable_order='
                        search_option += 'reverse_level)),shrink_strategy='
                        search_option += f'shrink_bisimulation(greedy={greedy}),'
                        search_option += 'label_reduction='
                        search_option += f'exact(before_shrinking={before_shrinking},'
                        search_option += f'before_merging={before_merging}),'
                        search_option += f'max_states={max_states}'

                        if max_states == 'true': 
                            threshold_before_merge += ',threshold_before_merge=1))'
                        else:
                            search_option += '))'
                    else:
                        search_option = config['fast_downward_search_config'] + '('
                        search_option += config['astar_h'] + '())'

                elif config['fast_downward_search_config'] in ('lazy_greedy', 'eager_greedy'):
                    h_def = ''
                    h_name = []
                    search_option = ''
                    skip = 0
                    for i in range(int(config['number_h'])):
                        idx = int(i) + 1
                        h = config[f'h_{idx}'] + '()'
                        if h in h_name:
                            skip += 1
                            continue
                        h_name.append(h)
                    search_option += config['fast_downward_search_config']
                    search_option += '(['
                    for i, hn in enumerate(h_name):
                        search_option += hn
                        if i + 1 < len(h_name):
                            search_option += ','
                        else:
                            search_option += '],'
                    search_option += 'preferred=['
                    prefs = ''
                    used = []
                    for i in range(int(config['nr_preferred'])):
                        idx = int(i) + 1
                        if idx <= int(config['number_h']):
                            h_idx = int(float(config[f'preferred_{idx}'])) - 1
                            if h_idx > len(h_name) - 1:
                                if len(h_name) > 1:
                                    h_idx = random.randint(0, len(h_name) - 1)
                                else:
                                    h_idx = 0
                            if h_name[h_idx] not in used:
                                prefs += h_name[h_idx]
                        if idx < len(h_name) and h_name[h_idx] not in used and idx != int(config['nr_preferred']):
                            prefs += ','
                        elif ']' not in prefs and idx == int(config['nr_preferred']):
                            if h_name[h_idx] in used and prefs[-1] == ',':
                                prefs = prefs[:-1]
                            prefs += '],'
                        used.append(h_name[h_idx])
                    search_option += prefs
                    cost_type = config['cost_type']
                    search_option += f'cost_type={cost_type})'

                elif config['fast_downward_search_config'] == 'lazy_wastar':
                    h_def = ''
                    h_name = []
                    search_option = ''
                    skip = 0
                    for i in range(int(config['number_h'])):
                        idx = int(i) + 1
                        h = config[f'h_{idx}'] + '()'
                        if h in h_name:
                            skip += 1
                            continue
                        h_name.append(h)
                    search_option += config['fast_downward_search_config']
                    search_option += '(['
                    for i, hn in enumerate(h_name):
                        search_option += hn
                        if i + 1 < len(h_name):
                            search_option += ','
                        else:
                            search_option += '],'
                    w = int(float(config['w']))
                    search_option += f'w={w},'
                    search_option += 'preferred=['
                    prefs = ''
                    used = []
                    for i in range(int(config['nr_preferred'])):
                        idx = int(i) + 1    
                        if idx <= int(config['number_h']):
                            h_idx = int(float(config[f'preferred_{idx}'])) - 1
                            if h_idx > len(h_name) - 1:
                                if len(h_name) > 1:
                                    h_idx = random.randint(0, len(h_name) - 1)
                                else:
                                    h_idx = 0
                            if h_name[h_idx] not in used not in used:
                                prefs += h_name[h_idx]
                        if idx < len(h_name) and h_name[h_idx] not in used and idx != int(config['nr_preferred']):
                            prefs += ','
                        elif ']' not in prefs and idx == int(config['nr_preferred']):
                            if h_name[h_idx] in used and prefs[-1] == ',':
                                prefs = prefs[:-1]
                            prefs += '],'
                        used.append(h_name[h_idx])
                    search_option += prefs
                    cost_type = config['cost_type']
                    search_option += f'cost_type={cost_type})'

                elif config['fast_downward_search_config'] == 'eager':
                    h = config['h_1'] + '()'
                    h_name = h
                    search_option = ''
                    weight = int(float(config['weight']))
                    search_option += config['fast_downward_search_config']
                    search_option += '(single(sum([g(),weight(' + f'{h_name}'
                    search_option += f',{weight})])),'
                    search_option += f'preferred=[{h_name}],'
                    cost_type = config['cost_type']
                    search_option += f'cost_type={cost_type})'

                if plantype == 'OneshotPlanner':
                    config = {'symk_search_config': search_option}

                elif plantype == 'AnytimePlanner':
                    config = {'symk_anytime_search_config': search_option}

        elif engine in ('tamer', 'pyperplan'):
            if isinstance(configuration, dict):
                config = configuration
            else:
                config = configuration.get_dictionary()

        elif engine in ('enhsp', 'enhsp-any'):
            if isinstance(configuration, dict):
                config = configuration
            else:
                config = configuration.get_dictionary()
            if not isinstance(config, str) and 'params' not in config:
                conf = ''
                for name, value in config.items():
                    conf += '-' + name[0] + ' ' + str(value) + ' '
                config = {'params': conf}

        return config



"""Functionalities for managing and calling selector."""
import os
import sys
import shutil
import json
from ConfigSpace.read_and_write import pcs

from selector.run_ac import ac

import signal
from contextlib import contextmanager

sys.path.append(os.path.abspath(os.path.dirname(__file__)))
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from AC_interface import *
from configurators import Configurator


class TimeoutException(Exception):
    pass


@contextmanager
def time_limit(seconds):
    def signal_handler(signum, frame):
        raise TimeoutException("Timed out!")
    signal.signal(signal.SIGALRM, signal_handler)
    signal.alarm(seconds)
    try:
        yield
    finally:
        signal.alarm(0)


class SelectorConfigurator(Configurator):
    """Configurator functions."""

    def __init__(self):
        """Initialize Selector configurator."""
        Configurator.__init__(self)
        self.crash_cost = 0
        self.planner_timelimit = 0
        self.engine = None
        self.gaci = None 

    def get_feedback_function(self, gaci, engine, metric, mode,
                              gray_box=False):
        """
        Generate a function to run the planning engine and obtain feedback.

        :param ACInterface gaci: Algorithm Configuration Interface object.
        :param str engine: Name of the planning engine.
        :param str metric: Metric to optimize, either 'runtime' or 'quality'.
        :param str mode: Type of planning.
        :param bool gray_box: True if using a gray box, False otherwise (optional).

        :return: A planner feedback function that takes configuration, instance, seed, and reader.
        :rtype: function

        :raises ValueError: If the provided engine is not supported for the given metric and mode.
        """

        if engine in self.capabilities[metric][mode]:
            self.metric = metric

            feedback_args = {'engine': engine, 'metric': metric,
                             'mode': mode, 'gray_box': gray_box,
                             'patience': self.patience,
                             'train_set': self.train_set,
                             'timelimit': self.scenario['cutoff_time'],
                             'crash_cost': self.crash_cost,
                             'verbose': self.verbose}

            fs_path = f'{self.feedback_path}/sel_feedback_setting'

            if not os.path.exists(fs_path):
                os.makedirs(fs_path)

            with open(f'{fs_path}/feedback_args.txt', 'w') as f:
                f.write(str(feedback_args))

            planner_feedback = f'{fs_path}/feedback_args.txt'

            return planner_feedback
        else:
            if self.verbose:
                print(f'''Algorithm Configuration for {metric} of {engine}
                      in {mode} is not supported.''')
            return None

    def set_scenario(self, engine, param_space, gaci,
                     configuration_time=120, n_trials=400, min_budget=1,
                     max_budget=3, crash_cost=0, planner_timelimit=30,
                     n_workers=1, instances=[], instance_features=None,
                     tourn_size=8, metric='runtime', memory_limit=2048,
                     patience=10, output_dir='up_ac_selector',
                     ray_mode='desktop'):
        """
        Set up the algorithm configuration scenario for Selector.

        :param str engine: The name of the planning engine.
        :param ConfigSpace.ConfigurationSpace param_space: ConfigSpace object defining the parameter space.
        :param ACInterface gaci: AC interface object.
        :param int configuration_time: Overall configuration time budget in seconds (default is 120).
        :param int n_trials: Maximum number of engine evaluations (default is 400).
        :param int min_budget: Minimum number of instances to use (default is 1).
        :param int max_budget: Maximum number of instances to use (default is 3).
        :param int crash_cost: The cost to use if the engine fails (default is 0).
        :param int planner_timelimit: Maximum runtime per evaluation for the planner (default is 30).
        :param int n_workers: Number of cores to utilize (default is 1).
        :param list instances: List of problem instance paths (default is empty list, uses train_set).
        :param dict instance_features: Dictionary containing instance names and lists of features (default is None).
        :param str metric: The optimization metric, either 'runtime' or 'quality' (default is 'runtime').
        :param int memory_limit: At which amount of MB used by a planning engine to terminate the run.
        :param int patience: Extra time in seconds to grant engine to terminate gracefully.
        :param str output_dir: Name of the directory selector writes to (directory is at .)
        :param str ray_mode: Set to 'desktop' if running locally, or 'cluster' if running in SLURM

        :raises ValueError: If an unsupported metric is provided.
        """
        if not instances:
            instances = self.train_set
        if isinstance(instances[0], tuple):
            train_set_dict = {}
            clean_inst = instances
            instances = []
            for inst in clean_inst:
                train_set_dict[inst[0]] = inst
                instances.append(inst[0])
            self.train_set = train_set_dict
        try:
            import up_ac
            path = '/' + os.path.abspath(up_ac.__file__).strip('/__init__.py')
            path += '/utils'
        except ImportError:
            path = os.getcwd().rsplit('up_ac', 1)[0]
            if path[-1] != "/":
                path += "/"
            path += 'up_ac/utils'

        self.feedback_path = path
        with open(f'{self.feedback_path}/instances.txt', 'w') as f:
            for inst in instances:
                f.write("%s\n" % inst)
        with open(f'{self.feedback_path}/features.txt', 'w') as f:
            header = ['inst_name'] + [str(feat_nr) 
                                      for feat_nr in range(
                                          len(instance_features[list(
                                              instance_features.keys())[0]]))]
            f.write("%s\n" % ', '.join(header))
            for inst, feats in instance_features.items():
                feats = map(str, feats)
                f.write("%s, %s\n" % (inst, ', '.join(feats)))
        with open(f'{self.feedback_path}/configspace.pcs', 'w') as f:
            f.write(pcs.write(param_space))
        if metric == 'runtime':
            if crash_cost != planner_timelimit:
                crash_cost = planner_timelimit
        self.crash_cost = crash_cost
        self.patience = patience
        self.metric = metric
        if self.metric == 'quality':
            self.planner_timelimit = planner_timelimit + patience
        else:
            self.planner_timelimit = planner_timelimit
        self.engine = engine
        self.gaci = gaci
        self.n_workers = n_workers
        number_tournaments = int(n_workers / tourn_size)
        try:
            import up_ac
            path = '/' + os.path.abspath(up_ac.__file__).strip('/__init__.py')
            path += '/utils'
        except ImportError:
            path = os.getcwd().rsplit('up_ac', 1)[0]
            if path[-1] != "/":
                path += "/"
            path += 'up_ac/utils'
        shutil.copyfile(f'{path}/up_ac_wrapper.py', f'up_ac_wrapper.py')
        if len(self.train_set) < 8:
            smac_pca_dim = len(self.train_set)
        else:
            smac_pca_dim = 8

        scen_files = {'paramfile': f'{self.feedback_path}/configspace.pcs',
                      'instance_file': f'{self.feedback_path}/instances.txt',
                      'feature_file': f'{self.feedback_path}/features.txt'}

        self.scenario = \
            dict(scen_files=scen_files, ray_mode=ray_mode,
                 run_obj=metric, overall_obj='PAR10',
                 cutoff_time=planner_timelimit, seed=44, par=10,
                 winners_per_tournament=1, tournament_size=tourn_size,
                 number_tournaments=number_tournaments,
                 termination_criterion='total_runtime',
                 monitor='tournament_level',
                 initial_instance_set_size=min_budget,
                 set_size=max_budget, generator_multiple=1,
                 memory_limit=memory_limit, 
                 check_path=False, log_folder=output_dir,
                 wallclock_limit=configuration_time,
                 wrapper_mod_name='up_ac_wrapper', deterministic=0,
                 cleanup=True, wrapper_class_name='UP_AC_Wrapper',
                 smac_pca_dim=smac_pca_dim, crash_cost=crash_cost,
                 cpu_binding=True)

        if metric == 'runtime':
            self.scenario['solve_match'] = r'Tuned plan runtime is:'
            self.scenario['runtime_feedback'] = \
                r"\s*([+-]?\d+(?:\.\d+)?)"

        elif metric == 'quality':
            self.scenario['quality_match'] = r'Tuned plan quality is:'
            self.scenario['quality_extract'] = r'[-+]?\d+'

        if self.verbose:
            print('\nSelector scenario is set.\n')

    def optimize(self, feedback_function=None, gray_box=False):
        """
        Run the algorithm configuration optimization.

        :param function feedback_function: A function to run the engine and obtain feedback (optional).
        :param bool gray_box: True if gray box usage is enabled, False otherwise (optional).

        :return: A tuple containing the best configuration found and additional information (if available).
                Returns None if feedback_function is not provided.
        :rtype: tuple or None
        """
        if feedback_function is not None:
            # Import feedback function, since dask cannot pickle local objects            
            path = os.getcwd().rsplit('up_ac', 1)[0]
            if path[-1] != "/":
                path += "/"
            path += 'up_ac/utils'
            sys.path.append(r"{}".format(path))

            if self.verbose:
                print('\nStarting Parameter optimization\n')

            __name__ = '__main__'

            ac(self.scenario.pop('scen_files'), self.scenario.pop('ray_mode'),
               **self.scenario)

            os.remove('up_ac_wrapper.py')

            lf = self.scenario['log_folder']

            with open(f'./selector/logs/{lf}/overall_best.json') as f:
                self.incumbent = json.load(f)

            self.incumbent = \
                self.incumbent[list(self.incumbent.keys())[0]]['conf']

            if self.verbose:
                print('\nBest Configuration found is:\n',
                      self.incumbent)

            return self.incumbent, None
        else:
            return None, None


"""Wrapper to call UP planning engines from selector."""

import argparse
import ast
import os


class UP_AC_Wrapper():

    def get_command_line_args(self, runargs, config):
        """Generate command line to call planner from UP."""

        instance = runargs["instance"]

        try:
            import up_ac
            path = '/' + os.path.abspath(up_ac.__file__).strip('/__init__.py')
            path += '/utils'
        except ImportError:
            path = os.getcwd().rsplit('up_ac', 1)[0]
            if path[-1] != "/":
                path += "/"
            path += 'up_ac/utils'

        exc = path + '/sel_call_up.py'

        cmd = f"stdbuf -oL python3 {exc} --i {instance} --c \"{config}\" "
        return cmd


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument('--runargs', type=ast.literal_eval)
    parser.add_argument('--config', type=ast.literal_eval)

    args = vars(parser.parse_args())

    wrapper = UP_AC_Wrapper()



from __future__ import annotations
from smac.runner.abstract_runner import AbstractRunner
from smac.facade.abstract_facade import AbstractFacade
from pathlib import Path
import dask
from dask.distributed import Client, Future, wait

from typing import Callable
import time

import joblib
from typing_extensions import Literal

from smac.acquisition.function.abstract_acquisition_function import (
    AbstractAcquisitionFunction,
)
from smac.acquisition.maximizer.abstract_acqusition_maximizer import (
    AbstractAcquisitionMaximizer,
)
from ConfigSpace import Configuration

from smac.callback.callback import Callback
from smac.initial_design.abstract_initial_design import AbstractInitialDesign
import smac.intensifier.abstract_intensifier as aiai  # import AbstractIntensifier
from smac.main.config_selector import ConfigSelector
from smac.model.abstract_model import AbstractModel
from smac.multi_objective.abstract_multi_objective_algorithm import (
    AbstractMultiObjectiveAlgorithm,
)
from smac.random_design.abstract_random_design import AbstractRandomDesign
from smac.runhistory.encoder.abstract_encoder import AbstractRunHistoryEncoder
from smac.runhistory.runhistory import RunHistory
from smac.runner.target_function_runner import TargetFunctionRunner
from smac.runner.target_function_script_runner import TargetFunctionScriptRunner
from smac.scenario import Scenario
from smac.utils.logging import get_logger, setup_logging

from smac.acquisition.function.expected_improvement import EI
from smac.acquisition.maximizer.local_and_random_search import (
    LocalAndSortedRandomSearch,
)
from smac.initial_design.default_design import DefaultInitialDesign
from smac.intensifier.intensifier import Intensifier
from smac.model.random_forest.random_forest import RandomForest
from smac.multi_objective.aggregation_strategy import MeanAggregationStrategy
from smac.random_design.probability_design import ProbabilityRandomDesign
from smac.runhistory.encoder.encoder import RunHistoryEncoder

from types import MethodType
import logging

logger = get_logger(__name__)

__copyright__ = "Copyright 2022, automl.org"
__license__ = "3-clause BSD"


def print_config_changes_new(
    incumbent: Configuration | None,
    challenger: Configuration | None,
    logger: logging.Logger,
) -> None:
    """Compares two configurations and prints the differences."""
    if incumbent is None or challenger is None:
        return

    params = sorted(
        [
            (param, incumbent[param], challenger[param])
            if param in incumbent.keys()
            else (param, None, challenger[param])
            for param in challenger.keys()
        ]
    )
    for param in params:
        if param[1] != param[2]:
            logger.info("--- %s: %r -> %r" % param)
        else:
            logger.debug("--- %s Remains unchanged: %r", param[0], param[1])


aiai.print_config_changes = print_config_changes_new


class DaskParallelRunner_new(AbstractRunner):

    """Interface to submit and collect a job in a distributed fashion. DaskParallelRunner is
    intended to comply with the bridge design pattern. Nevertheless, to reduce the amount of code
    within single-vs-parallel implementations, DaskParallelRunner wraps a BaseRunner object which
    is then executed in parallel on `n_workers`.

    This class then is constructed by passing an AbstractRunner that implements
    a `run` method, and is capable of doing so in a serial fashion. Next,
    this wrapper class uses dask to initialize `N` number of AbstractRunner that actively wait of a
    TrialInfo to produce a RunInfo object.

    To be more precise, the work model is then:

    1. The intensifier dictates "what" to run (a configuration/instance/seed) via a TrialInfo object.
    2. An abstract runner takes this TrialInfo object and launches the task via
       `submit_run`. In the case of DaskParallelRunner, `n_workers` receive a pickle-object of
       `DaskParallelRunner.single_worker`, each with a `run` method coming from
       `DaskParallelRunner.single_worker.run()`
    3. TrialInfo objects are run in a distributed fashion, and their results are available locally to each worker. The
       result is collected by `iter_results` and then passed to SMBO.
    4. Exceptions are also locally available to each worker and need to be collected.

    Dask works with `Future` object which are managed via the DaskParallelRunner.client.

    Parameters
    ----------
    single_worker : AbstractRunner
        A runner to run in a distributed fashion. Will be distributed using `n_workers`.
    patience: int, default to 5
        How much to wait for workers (seconds) to be available if one fails.
    dask_client: Client | None, defaults to None
        User-created dask client, which can be used to start a dask cluster and then attach SMAC to it. This will not
        be closed automatically and will have to be closed manually if provided explicitly. If none is provided
        (default), a local one will be created for you and closed upon completion.
    """

    def __init__(
        self,
        single_worker: AbstractRunner,
        patience: int = 5,
        dask_client: Client | None = None,
    ):
        super().__init__(
            scenario=single_worker._scenario,
            required_arguments=single_worker._required_arguments,
        )

        # The single worker to hold on to and call run on
        self._single_worker = single_worker

        # The list of futures that dask will use to indicate in progress runs
        self._pending_trials: list[Future] = []

        # Dask related variables
        self._scheduler_file: Path | None = None
        self._patience = patience

        self._client: Client
        self._close_client_at_del: bool

        if dask_client is None:
            dask.config.set({"distributed.worker.daemon": False})
            self._close_client_at_del = True
            self._client = Client(
                n_workers=self._scenario.n_workers,
                processes=True,
                threads_per_worker=1,
                local_directory=str(self._scenario.output_directory),
                timeout=300
            )

            if self._scenario.output_directory is not None:
                self._scheduler_file = Path(self._scenario.output_directory, ".dask_scheduler_file")
                self._client.write_scheduler_file(scheduler_file=str(self._scheduler_file))
        else:
            # We just use their set up
            self._client = dask_client
            self._close_client_at_del = False

    def submit_trial(self, trial_info: TrialInfo, **dask_data_to_scatter: dict[str, Any]) -> None:
        """This function submits a configuration embedded in a ``trial_info`` object, and uses one of
        the workers to produce a result locally to each worker.

        The execution of a configuration follows this procedure:

        #. The SMBO/intensifier generates a `TrialInfo`.
        #. SMBO calls `submit_trial` so that a worker launches the `trial_info`.
        #. `submit_trial` internally calls ``self.run()``. It does so via a call to `run_wrapper` which contains common
           code that any `run` method will otherwise have to implement.

        All results will be only available locally to each worker, so the main node needs to collect them.

        Parameters
        ----------
        trial_info : TrialInfo
            An object containing the configuration launched.

        dask_data_to_scatter: dict[str, Any]
            When a user scatters data from their local process to the distributed network,
            this data is distributed in a round-robin fashion grouping by number of cores.
            Roughly speaking, we can keep this data in memory and then we do not have to (de-)serialize the data
            every time we would like to execute a target function with a big dataset.
            For example, when your target function has a big dataset shared across all the target function,
            this argument is very useful.
        """
        # Check for resources or block till one is available
        if self.count_available_workers() <= 0:
            logger.debug("No worker available. Waiting for one to be available...")
            wait(self._pending_trials, return_when="FIRST_COMPLETED")
            self._process_pending_trials()

        # Check again to make sure that there are resources
        if self.count_available_workers() <= 0:
            logger.warning("No workers are available. This could mean workers crashed. Waiting for new workers...")
            time.sleep(self._patience)
            if self.count_available_workers() <= 0:
                raise RuntimeError(
                    "Tried to execute a job, but no worker was ever available."
                    "This likely means that a worker crashed or no workers were properly configured."
                )

        # At this point we can submit the job
        trial = self._client.submit(self._single_worker.run_wrapper, trial_info=trial_info, **dask_data_to_scatter)
        self._pending_trials.append(trial)

    def iter_results(self) -> Iterator[tuple[TrialInfo, TrialValue]]:  # noqa: D102
        self._process_pending_trials()
        while self._results_queue:
            yield self._results_queue.pop(0)

    def wait(self) -> None:  # noqa: D102
        if self.is_running():
            wait(self._pending_trials, return_when="FIRST_COMPLETED")

    def is_running(self) -> bool:  # noqa: D102
        return len(self._pending_trials) > 0

    def run(
        self,
        config: Configuration,
        instance: str | None = None,
        budget: float | None = None,
        seed: int | None = None,
        **dask_data_to_scatter: dict[str, Any],
    ) -> tuple[StatusType, float | list[float], float, dict]:  # noqa: D102
        return self._single_worker.run(
            config=config, instance=instance, seed=seed, budget=budget, **dask_data_to_scatter
        )

    def count_available_workers(self) -> int:
        """Total number of workers available. This number is dynamic as more resources
        can be allocated.
        """
        return sum(self._client.nthreads().values()) - len(self._pending_trials)

    def close(self, force: bool = False) -> None:
        """Closes the client."""
        if self._close_client_at_del or force:
            self._client.close()

    def _process_pending_trials(self) -> None:
        """The completed trials are moved from ``self._pending_trials`` to ``self._results_queue``.
        We make sure pending trials never exceed the capacity of the scheduler.
        """
        # In code check to make sure we don't exceed resource allocation
        if self.count_available_workers() < 0:
            logger.warning(
                "More running jobs than resources available. "
                "Should not have more pending trials in remote workers "
                "than the number of workers. This could mean a worker "
                "crashed and was not able to be recovered by dask. "
            )

        # Move the done run from the worker to the results queue
        done = [trial for trial in self._pending_trials if trial.done()]
        for trial in done:
            self._results_queue.append(trial.result())
            self._pending_trials.remove(trial)

    def __del__(self) -> None:
        """Makes sure that when this object gets deleted, the client is terminated. This
        is only done if the client was created by the dask runner.
        """
        if self._close_client_at_del:
            self.close()


class AbstractFacade_new(AbstractFacade):

    def __init__(
        self,
        scenario: Scenario,
        target_function: Callable | str | AbstractRunner,
        *,
        model: AbstractModel | None = None,
        acquisition_function: AbstractAcquisitionFunction | None = None,
        acquisition_maximizer: AbstractAcquisitionMaximizer | None = None,
        initial_design: AbstractInitialDesign | None = None,
        random_design: AbstractRandomDesign | None = None,
        intensifier: aiai.AbstractIntensifier | None = None,
        multi_objective_algorithm: AbstractMultiObjectiveAlgorithm | None = None,
        runhistory_encoder: AbstractRunHistoryEncoder | None = None,
        config_selector: ConfigSelector | None = None,
        logging_level: int | Path | Literal[False] | None = None,
        callbacks: list[Callback] = [],
        overwrite: bool = False,
        dask_client: Client | None = None,
    ):
        setup_logging(logging_level)

        if model is None:
            model = self.get_model(scenario)

        if acquisition_function is None:
            acquisition_function = self.get_acquisition_function(scenario)

        if acquisition_maximizer is None:
            acquisition_maximizer = self.get_acquisition_maximizer(scenario)

        if initial_design is None:
            initial_design = self.get_initial_design(scenario)

        if random_design is None:
            random_design = self.get_random_design(scenario)

        if intensifier is None:
            intensifier = self.get_intensifier(scenario)

        if multi_objective_algorithm is None and scenario.count_objectives() > 1:
            multi_objective_algorithm = self.get_multi_objective_algorithm(scenario=scenario)

        if runhistory_encoder is None:
            runhistory_encoder = self.get_runhistory_encoder(scenario)

        if config_selector is None:
            config_selector = self.get_config_selector(scenario)

        # Initialize empty stats and runhistory object
        runhistory = RunHistory(multi_objective_algorithm=multi_objective_algorithm)

        # Set the seed for configuration space
        scenario.configspace.seed(scenario.seed)

        # Set variables globally
        self._scenario = scenario
        self._model = model
        self._acquisition_function = acquisition_function
        self._acquisition_maximizer = acquisition_maximizer
        self._initial_design = initial_design
        self._random_design = random_design
        self._intensifier = intensifier
        self._multi_objective_algorithm = multi_objective_algorithm
        self._runhistory = runhistory
        self._runhistory_encoder = runhistory_encoder
        self._config_selector = config_selector
        self._callbacks = callbacks
        self._overwrite = overwrite

        # Prepare the algorithm executer
        runner: AbstractRunner
        if isinstance(target_function, AbstractRunner):
            runner = target_function
        elif isinstance(target_function, str):
            runner = TargetFunctionScriptRunner(
                scenario=scenario,
                target_function=target_function,
                required_arguments=self._get_signature_arguments(),
            )
        else:
            runner = TargetFunctionRunner(
                scenario=scenario,
                target_function=target_function,
                required_arguments=self._get_signature_arguments(),
            )

        # In case of multiple jobs, we need to wrap the runner again using DaskParallelRunner
        if (n_workers := scenario.n_workers) > 1 or dask_client is not None:
            if dask_client is not None:
                logger.warning(
                    "Provided `dask_client`. Ignore `scenario.n_workers`, directly set `n_workers` in `dask_client`."
                )
            else:
                available_workers = joblib.cpu_count()
                if n_workers > available_workers:
                    logger.info(f"Workers are reduced to {available_workers}.")
                    n_workers = available_workers

            # We use a dask runner for parallelization
            runner = DaskParallelRunner_new(single_worker=runner, dask_client=dask_client)

        # Set the runner to access it globally
        self._runner = runner

        # Adding dependencies of the components
        self._update_dependencies()

        # We have to update our meta data (basically arguments of the components)
        self._scenario._set_meta(self.meta)

        # We have to validate if the object compositions are correct and actually make sense
        self._validate()

        # Finally we configure our optimizer
        self._optimizer = self._get_optimizer()
        assert self._optimizer

        # Register callbacks here
        for callback in callbacks:
            self._optimizer.register_callback(callback)

        # Additionally, we register the runhistory callback from the intensifier to efficiently update our incumbent
        # every time new information are available
        self._optimizer.register_callback(self._intensifier.get_callback(), index=0)


class AlgorithmConfigurationFacade_new(AbstractFacade_new):
    @staticmethod
    def get_model(  # type: ignore
        scenario: Scenario,
        *,
        n_trees: int = 10,
        ratio_features: float = 5.0 / 6.0,
        min_samples_split: int = 3,
        min_samples_leaf: int = 3,
        max_depth: int = 20,
        bootstrapping: bool = True,
        pca_components: int = 4,
    ) -> RandomForest:
        """Returns a random forest as surrogate model.

        Parameters
        ----------
        n_trees : int, defaults to 10
            The number of trees in the random forest.
        ratio_features : float, defaults to 5.0 / 6.0
            The ratio of features that are considered for splitting.
        min_samples_split : int, defaults to 3
            The minimum number of data points to perform a split.
        min_samples_leaf : int, defaults to 3
            The minimum number of data points in a leaf.
        max_depth : int, defaults to 20
            The maximum depth of a single tree.
        bootstrapping : bool, defaults to True
            Enables bootstrapping.
        pca_components : float, defaults to 4
            Number of components to keep when using PCA to reduce dimensionality of instance features.
        """
        return RandomForest(
            configspace=scenario.configspace,
            n_trees=n_trees,
            ratio_features=ratio_features,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            max_depth=max_depth,
            bootstrapping=bootstrapping,
            log_y=False,
            instance_features=scenario.instance_features,
            pca_components=pca_components,
            seed=scenario.seed,
        )

    @staticmethod
    def get_acquisition_function(  # type: ignore
        scenario: Scenario,
        *,
        xi: float = 0.0,
    ) -> EI:
        """Returns an Expected Improvement acquisition function.

        Parameters
        ----------
        scenario : Scenario
        xi : float, defaults to 0.0
            Controls the balance between exploration and exploitation of the
            acquisition function.
        """
        return EI(xi=xi)

    @staticmethod
    def get_acquisition_maximizer(  # type: ignore
        scenario: Scenario,
    ) -> LocalAndSortedRandomSearch:
        """Returns local and sorted random search as acquisition maximizer."""
        optimizer = LocalAndSortedRandomSearch(
            scenario.configspace,
            seed=scenario.seed,
        )

        return optimizer

    @staticmethod
    def get_intensifier(
        scenario: Scenario,
        *,
        max_config_calls: int = 2000,
        max_incumbents: int = 10,
    ) -> Intensifier:
        """Returns ``Intensifier`` as intensifier. Supports budgets.

        Parameters
        ----------
        max_config_calls : int, defaults to 3
            Maximum number of configuration evaluations. Basically, how many instance-seed keys should be evaluated at
            maximum for a configuration.
        max_incumbents : int, defaults to 10
            How many incumbents to keep track of in the case of multi-objective.
        """
        return Intensifier(
            scenario=scenario,
            max_config_calls=max_config_calls,
            max_incumbents=max_incumbents,
        )

    @staticmethod
    def get_initial_design(  # type: ignore
        scenario: Scenario,
        *,
        additional_configs: list[Configuration] = [],
    ) -> DefaultInitialDesign:
        """Returns an initial design, which returns the default configuration.

        Parameters
        ----------
        additional_configs: list[Configuration], defaults to []
            Adds additional configurations to the initial design.
        """
        return DefaultInitialDesign(
            scenario=scenario,
            additional_configs=additional_configs,
        )

    @staticmethod
    def get_random_design(  # type: ignore
        scenario: Scenario,
        *,
        probability: float = 0.5,
    ) -> ProbabilityRandomDesign:
        """Returns ``ProbabilityRandomDesign`` for interleaving configurations.

        Parameters
        ----------
        probability : float, defaults to 0.5
            Probability that a configuration will be drawn at random.
        """
        return ProbabilityRandomDesign(probability=probability, seed=scenario.seed)

    @staticmethod
    def get_multi_objective_algorithm(  # type: ignore
        scenario: Scenario,
        *,
        objective_weights: list[float] | None = None,
    ) -> MeanAggregationStrategy:
        """Returns the mean aggregation strategy for the multi objective algorithm.

        Parameters
        ----------
        scenario : Scenario
        objective_weights : list[float] | None, defaults to None
            Weights for averaging the objectives in a weighted manner. Must be of the same length as the number of
            objectives.
        """
        return MeanAggregationStrategy(
            scenario=scenario,
            objective_weights=objective_weights,
        )

    @staticmethod
    def get_runhistory_encoder(scenario: Scenario) -> RunHistoryEncoder:
        """Returns the default runhistory encoder."""
        return RunHistoryEncoder(scenario)


def patch_daskrunner(facade):

    return AlgorithmConfigurationFacade_new



"""Functionalities for managing and calling configurators."""
from irace import irace
from unified_planning.exceptions import UPProblemDefinitionError, UPException

import timeit
import signal
import sys
import os
from contextlib import contextmanager
from pebble import concurrent
from pebble.common import ProcessExpired
from concurrent.futures import TimeoutError

sys.path.append(os.path.abspath(os.path.dirname(__file__)))
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from AC_interface import *
from configurators import Configurator


class TimeoutException(Exception):
    pass


@contextmanager
def time_limit(seconds):
    def signal_handler(signum, frame):
        raise TimeoutException("Timed out!")
    signal.signal(signal.SIGALRM, signal_handler)
    signal.alarm(seconds)
    try:
        yield
    finally:
        signal.alarm(0)


class IraceConfigurator(Configurator):
    """Configurator functions."""

    def __init__(self):
        """Initialize Irace configurator."""
        Configurator.__init__(self)

    def get_feedback_function(self, gaci, engine, metric, mode,
                              gray_box=False):
        """
        Generate a function to run the planning engine and obtain feedback.

        :param gaci: AC interface object.
        :type gaci: object
        :param engine: Name of the planning engine.
        :type engine: str
        :param metric: Metric type, either 'runtime' or 'quality'.
        :type metric: str
        :param mode: Type of planning mode.
        :type mode: str
        :param gray_box: True if using a gray box approach (optional).
        :type gray_box: bool, optional

        :return: A function to provide feedback based on the specified parameters.
        :rtype: function

        :raises ValueError: If the provided engine, metric, or mode is not supported.
        """
        if engine in self.capabilities[metric][mode]:
            self.metric = metric

            def planner_feedback(experiment, scenario):

                if metric == 'quality':
                    timelimit = self.planner_timelimit - self.patience
                else:
                    timelimit = self.planner_timelimit

                start = timeit.default_timer()
                instance_p = \
                    self.scenario['instances'][experiment['id.instance'] - 1]
                try:
                    if ',' in instance_p:
                        instance_p = eval(instance_p)
                        pddl_problem = \
                            self.reader.parse_problem(f'{instance_p[1]}',
                                                      f'{instance_p[0]}')
                    else:
                        domain_path = instance_p.rsplit('/', 1)[0]
                        domain = f'{domain_path}/domain.pddl'
                        pddl_problem = self.reader.parse_problem(f'{domain}',
                                                                 f'{instance_p}')
                    config = dict(experiment['configuration'])

                except UPProblemDefinitionError as e:
                    print(e)
                    if metric == 'runtime':
                        feedback = timelimit * 2
                        feedback = {'cost': feedback, 'time': feedback}
                    else:
                        feedback = self.crash_cost
                        feedback = {'cost': feedback, 'time': timelimit * 2}
                    return feedback

                '''
                feedback = \
                    gaci.run_engine_config(config,
                                           metric,
                                           engine,
                                           mode,
                                           pddl_problem)
                '''
                
                try:
                    if engine == 'tamer' or engine == 'pyperplan':
                        @concurrent.process(timeout=timelimit)
                        def solve(config, metric, engine,
                                  mode, pddl_problem):
                            feedback = \
                                gaci.run_engine_config(config,
                                                       metric, engine,
                                                       mode, pddl_problem,
                                                       timelimit)

                            return feedback
                    else:
                        def solve(config, metric, engine,
                                  mode, pddl_problem):
                            feedback = \
                                gaci.run_engine_config(config,
                                                       metric, engine,
                                                       mode, pddl_problem,
                                                       timelimit)

                            return feedback

                    try:
                        if engine == 'tamer' or engine == 'pyperplan':
                            future = solve(config, metric, engine,
                                           mode, pddl_problem)
                            if metric == 'runtime':
                                penalty = timelimit
                            elif metric == 'quality':
                                penalty = self.crash_cost
                            try:
                                feedback = future.result()
                            except (TimeoutError, ProcessExpired) as err:
                                print(err)
                                feedback = penalty
                            except Exception as e:
                                print(e)
                                feedback = penalty
                        else:
                            with time_limit(timelimit):
                                feedback = solve(config, metric, engine,
                                                 mode, pddl_problem)

                    except TimeoutException:  # TimeoutError:
                        if metric == 'runtime':
                            feedback = timelimit
                        elif metric == 'quality':
                            feedback = self.crash_cost

                except (AssertionError, NotImplementedError,
                        UPProblemDefinitionError, UPException,
                        UnicodeDecodeError) as err:
                    if self.verbose:
                        print('\n** Error in planning engine!', err)
                    if metric == 'runtime':
                        feedback = timelimit
                    elif metric == 'quality':
                        feedback = self.crash_cost

                if feedback == 'unsolvable':
                    print('\n Unsolvable feedback:', feedback, '\n')
                    return {'cost': (self.crash_cost), 'time': timelimit}

                elif feedback is not None:
                    if metric == 'quality':
                        self.print_feedback(engine, instance_p, feedback)
                        runtime = timeit.default_timer() - start
                        feedback = {'cost': feedback, 'time': runtime}
                        return feedback
                    elif metric == 'runtime':
                        if engine in ('tamer', 'pyperplan') and feedback == 'measure':
                            feedback = timeit.default_timer() - start
                        self.print_feedback(engine, instance_p, feedback)
                        feedback = {'cost': feedback, 'time': feedback}
                        return feedback
                else:
                    if metric == 'runtime':
                        # Penalty is max runtime in runtime scenario
                        feedback = timelimit * 2
                        self.print_feedback(engine, instance_p, feedback)
                        feedback = {'cost': feedback, 'time': feedback}
                    else:
                        # Penalty is defined by user in quality scenario
                        feedback = self.crash_cost
                        self.print_feedback(engine, instance_p, feedback)
                        feedback = {'cost': feedback, 'time': timelimit * 2}
                    return feedback

            return planner_feedback
        else:
            if self.verbose:
                print(f'Algorithm Configuration for {metric} of {engine} in' + 
                      f' {mode} is not supported.')
            return None

    def set_scenario(self, engine, param_space, gaci,
                     configuration_time=120,
                     n_trials=400, min_budget=1, max_budget=3, crash_cost=0,
                     planner_timelimit=30, n_workers=1, instances=[],
                     instance_features=None, metric='runtime', patience=10):
        """
        Set up the algorithm configuration scenario.

        :param engine: Name of the engine.
        :type engine: str
        :param param_space: The ConfigSpace object defining the parameter space.
        :type param_space: ConfigSpace.ConfigurationSpace
        :param gaci: AC interface object.
        :type gaci: object
        :param configuration_time: Overall configuration time budget (optional).
        :type configuration_time: int, optional
        :param n_trials: Maximum number of engine evaluations (optional).
        :type n_trials: int, optional
        :param min_budget: Minimum number of instances to use (optional).
        :type min_budget: int, optional
        :param max_budget: Maximum number of instances to use (optional).
        :type max_budget: int, optional
        :param crash_cost: The cost to use if the engine fails (optional).
        :type crash_cost: int, optional
        :param planner_timelimit: Maximum runtime per evaluation (optional).
        :type planner_timelimit: int, optional
        :param n_workers: Number of cores to utilize (optional).
        :type n_workers: int, optional
        :param instances: List of problem instance paths (optional).
        :type instances: list, optional
        :param instance_features: Dictionary containing instance names and lists of features (optional).
        :type instance_features: dict, optional
        :param metric: Optimization metric, either 'runtime' or 'quality' (optional).
        :type metric: str, optional

        :raises ValueError: If the provided metric is not supported.
        """
        if not instances:
            if isinstance(self.train_set[0], tuple):
                instances = []
                for ts in self.train_set:
                    instances.append(str(ts))
            else:
                instances = self.train_set
        self.crash_cost = crash_cost
        self.patience = patience
        
        default_conf, forbiddens = gaci.get_ps_irace(param_space)

        if metric == 'quality':
            test_type = 'friedman'
            capping = False
            self.planner_timelimit = planner_timelimit + patience
        elif metric == 'runtime':
            test_type = 't-test'
            capping = True
            self.planner_timelimit = planner_timelimit
        # https://mlopez-ibanez.github.io/irace/reference/defaultScenario.html
        if forbiddens:
            scenario = dict(
                # We want to optimize for <configuration_time> seconds
                maxTime=configuration_time,  
                instances=instances,
                # List of training instances
                debugLevel=3, 
                # number of decimal places to be considered for real parameters
                digits=10, 
                # Number of parallel runs
                parallel=n_workers, 
                forbiddenFile="forbidden.txt",
                logFile="",
                initConfigurations=default_conf,
                # nbConfigurations=8,  # we let Irace decide
                deterministic=True,
                testType=test_type,
                capping=capping,
                boundMax=self.planner_timelimit,
                firstTest=min_budget
            )
        else:
            scenario = dict(
                # We want to optimize for <configuration_time> seconds
                maxTime=configuration_time, 
                # List of training instances
                instances=instances, 
                debugLevel=3, 
                # number of decimal places to be considered for real parameters
                digits=10, 
                # Number of parallel runs
                parallel=n_workers, 
                logFile="",
                initConfigurations=default_conf,
                # nbConfigurations=8,  # we let Irace decide
                deterministic=True,
                testType=test_type,
                capping=capping,
                boundMax=self.planner_timelimit,
                firstTest=min_budget
            )

        self.irace_param_space = gaci.irace_param_space

        if self.verbose:
            print('\nIrace scenario is set.\n')

        self.scenario = scenario

    def optimize(self, feedback_function=None, gray_box=False):
        """
        Run the algorithm configuration process.

        :param feedback_function: A function to run the engine and obtain feedback (optional).
        :type feedback_function: function, optional
        :param gray_box: True if using a gray box approach (optional).
        :type gray_box: bool, optional

        :returns: A tuple containing:
            - dict: The best configuration found.
            - None: If there is no feedback function.
        :rtype: tuple or None
        """

        if feedback_function is not None:

            if self.verbose:
                print('\nStarting Parameter optimization\n')
            ac = irace(self.scenario,
                       self.irace_param_space,
                       feedback_function)
            self.incumbent = ac.run()

            self.incumbent = self.incumbent.to_dict(orient='records')[0]

            if self.verbose:
                print('\nBest Configuration found is:\n',
                      self.incumbent)

            return self.incumbent, None
        else:
            return None, None



"""Functionalities for managing and calling configurators."""
from unified_planning.io import PDDLReader
from unified_planning.exceptions import UPProblemDefinitionError

import json
import timeit
import time
import os
import sys
import subprocess
import psutil
import signal
import multiprocessing

# make sure test can be run from anywhere
path = os.getcwd().rsplit('up-ac', 1)[0]
if path[-1] != "/":
    path += "/"
path += 'up-ac/up_ac'
if not os.path.isfile(sys.path[0] + '/configurators.py') and \
        'up-ac' in sys.path[0]:
    sys.path.append(sys.path[0].rsplit('up-ac', 1)[0] + 'up-ac')
    sys.path.append(sys.path[0].rsplit('up-ac', 1)[0] + 'up-ac/up_ac')
    sys.path.append(sys.path[0].rsplit('up-ac', 1)[0] + 'up-ac/utils')

sys.path.append(os.path.abspath(os.path.dirname(__file__)))
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from AC_interface import *


class Configurator():
    """Configurator functions."""

    def __init__(self):
        """Initialize generic interface."""
        self.capabilities = {'quality': {
                             'OneshotPlanner': 
                             ['lpg', 'fast-downward', 'enhsp', 'symk',
                              'pyperplan', 'tamer'],
                             'AnytimePlanner':
                             ['fast-downward', 'symk', 'enhsp-any',
                              'lpg-anytime']},
                             'runtime': {
                             'OneshotPlanner':
                             ['lpg', 'fast-downward', 'enhsp', 'symk',
                              'tamer', 'pyperplan']}
                             }
        self.incumbent = None
        self.instance_features = {}
        self.train_set = {}
        self.test_set = {}
        self.reader = PDDLReader()
        self.reader._env.error_used_name = False
        self.metric = None
        self.crash_cost = 0
        self.ac = None
        self.verbose = True

        # Control for integrated planner versions
        '''
        try:
            import up_enhsp
            cai('up-enhsp', '0.0.25')
        except (ImportError, ModuleNotFoundError):
            pass
        try:
            import up_fast_downward
            cai('up-fast-downward', '0.4.1')
        except (ImportError, ModuleNotFoundError):
            pass
        try:
            import up_lpg
            cai('up-lpg', '0.0.11')
        except (ImportError, ModuleNotFoundError):
            pass
        try:
            import up_pyperplan
            cai('up-pyperplan', '1.1.0')
        except (ImportError, ModuleNotFoundError):
            pass
        try:
            import up_symk
            cai('up-symk', '1.3.1')
        except (ImportError, ModuleNotFoundError):
            pass
        try:
            import up_tamer
            cai('up-tamer', '1.1.3')
        except (ImportError, ModuleNotFoundError):
            pass
        '''

    def print_feedback(self, engine, instance, feedback):
        """
        Print feedback from the engine.

        :param engine: Name of the engine.
        :type engine: str
        :param instance: Name of the instance.
        :type instance: str
        :param feedback: Feedback from the engine.
        """
        if self.verbose:
            print(f'** Feedback of {engine} on instance\n**' +
                  f' {instance}\n** is {feedback}\n\n')

    def get_instance_features(self, instance_features=None):
        """
        Save instance features.

        :param instance_features: Instance names and their features in lists.
        :type instance_features: dict, optional
        """
        self.instance_features = instance_features
        if self.verbose:
            print('\nSetting instance features.\n')

    def set_training_instance_set(self, train_set):
        """
        Save training instance set.

        :param train_set: List of instance paths.
        :type train_set: list
        """
        self.train_set = train_set
        if self.verbose:
            print('\nSetting training instance set.\n')

    def set_test_instance_set(self, test_set):
        """
        Set test instance set.

        :param test_set: List of instance paths.
        :type test_set: list
        """
        self.test_set = test_set
        if self.verbose:
            print('\nSetting testing instance set.\n')

    def get_feedback_function(self, gaci, engine, metric, mode,
                              gray_box=False):
        """
        Generate the function to run the engine and get feedback.

        :param gaci: Algorithm Configuration interface object.
        :type gaci: ACInterface
        :param engine: Engine name.
        :type engine: str
        :param metric: Metric, either 'runtime' or 'quality'.
        :type metric: str
        :param mode: Type of planning.
        :type mode: str
        :param gray_box: True if gray box to be used, optional.
        :type gray_box: bool, optional

        :return: Planner feedback function or None if not supported.
        :rtype: function or None
        """
        if engine in self.capabilities[metric][mode]:
            self.metric = metric

            planner_feedback = None

            return planner_feedback
        else:
            if self.verbose:
                print(f'Algorithm Configuration for {metric} of {engine}' + \
                      f' in {mode} is not supported.')
            return None

    def set_scenario(self, engine, param_space, gaci,
                     configuration_time=120, n_trials=400, min_budget=1,
                     max_budget=3, crash_cost=0, planner_timelimit=30,
                     n_workers=1, instances=[], instance_features=None,
                     metric='runtime', popSize=128, evalLimit=2147483647):
        """
        Set up algorithm configuration scenario.

        :param engine: Engine name.
        :type engine: str
        :param param_space: ConfigSpace object.
        :type param_space: ConfigSpace
        :param gaci: AC interface object.
        :type gaci: ACInterface
        :param configuration_time: Overall configuration time budget, optional.
        :type configuration_time: int, optional
        :param n_trials: Maximum number of engine evaluations, optional.
        :type n_trials: int, optional
        :param min_budget: Minimum number of instances to use, optional.
        :type min_budget: int, optional
        :param max_budget: Maximum number of instances to use, optional.
        :type max_budget: int, optional
        :param crash_cost: Cost to use if the engine fails, optional.
        :type crash_cost: int, optional
        :param planner_timelimit: Maximum runtime per evaluation, optional.
        :type planner_timelimit: int, optional
        :param n_workers: Number of cores to utilize, optional.
        :type n_workers: int, optional
        :param instances: Problem instance paths, optional.
        :type instances: list, optional
        :param instance_features: Instance names and lists of features, optional.
        :type instance_features: dict, optional
        :param metric: Optimization metric, optional.
        :type metric: str, optional
        :param popSize: Population size of configs per generation (OAT), optional.
        :type popSize: int, optional
        :param evalLimit: Maximum number of evaluations (OAT), optional.
        :type evalLimit: int, optional
        """
        self.planner_timelimit = planner_timelimit

        scenario = None

        self.scenario = scenario

    def optimize(self, feedback_function=None, gray_box=False):
        """
        Run the algorithm configuration.

        :param feedback_function: Function to run engine and get feedback, optional.
        :type feedback_function: function, optional
        :param gray_box: True if gray box usage, optional.
        :type gray_box: bool, optional

        :return: The best configuration found during optimization.
        :rtype: dict
        """
        if feedback_function is not None:

            return self.incumbent

    def get_process_by_name(self, names):
        """
        Finds all processes matching given names.

        :param names: List of names of processes to get.
        :type names: str

        :return: Processes that match the names.
        :rtype: list of psuitl.Process
        """
        matching_processes = []
        for proc in psutil.process_iter(attrs=['pid', 'name']):
            try:
                if any(proc.info['name'] in names for name in names):
                    matching_processes.append(proc)
            except (psutil.NoSuchProcess, psutil.AccessDenied,
                    psutil.ZombieProcess):
                continue
        return matching_processes

    def trigger_gc_for_children(self, pid):
        """
        Triggers garbage collection for child processes.

        :param pid: PID of parent Process.
        :type names: int
        """
        java_gcc_processes = self.get_process_by_name(["java", "gcc", "g++"])
        parent = psutil.Process(pid)
        for child in parent.children(recursive=True):
            try:
                process_name = child.name().lower()
                pid = child.pid

                if "java" in process_name:
                    print(f"Triggering GC for Java process PID: {pid}",
                          flush=True)
                    subprocess.run(["jcmd", str(pid), "GC.run"])

                elif "gcc" in process_name or "g++" in process_name:
                    print(f"Triggering GC for GCC process PID: {pid}",
                          flush=True)
                    # Sending SIGHUP as a best-effort attempt
                    child.send_signal(signal.SIGHUP)

            except Exception as e:
                print(f"Error triggering GC for process PID {pid}: {e}",
                      flush=True)

        # Monitor memory usage and wait until it reduces
        while any(proc.is_running() for proc in java_gcc_processes):
            try:
                while any(proc.memory_info().rss > 500 * 1024 * 1024 
                          for proc in java_gcc_processes):
                    print("Memory still high, waiting...")
                    time.sleep(5)
            except Exception as e:
                print(e)
                pass

    def kill_process_tree(self, pid, wt):
        """
        Terminate/kill all child processes.

        :param pid: PID of parent Process.
        :type names: int
        :param wt: Time to wait for process to terminate.
        :type names: int
        """
        try:
            parent = psutil.Process(pid)
            children = parent.children(recursive=True)
            for child in children:
                child.terminate()
            parent.terminate()

            # Wait for termination
            _, still_alive = psutil.wait_procs(children, timeout=wt)
            for p in still_alive:
                p.kill()  # Force kill if still alive

            parent.wait(wt)
        except (psutil.NoSuchProcess, psutil.TimeoutExpired) as err:
            print(err)

    def evaluate(self, metric, engine, mode, incumbent, gaci,
                 planner_timelimit=300, crash_cost=10000, instances=[]):
        """
        Evaluate performance of found configuration on training set.

        :param metric: Optimization metric.
        :type metric: str
        :param engine: Engine name.
        :type engine: str
        :param mode: Planning mode.
        :type mode: str
        :param incumbent: Parameter configuration to evaluate.
        :type incumbent: dict
        :param gaci: AC interface object.
        :type gaci: ACInterface
        :param planner_timelimit: Max runtime per evaluation, optional.
        :type planner_timelimit: int, optional
        :param crash_cost: Cost if engine fails, optional.
        :type crash_cost: int, optional
        :param instances: Instance paths, optional.
        :type instances: list, optional

        :return: Average performance on the instances.
        :rtype: float
        """
        if incumbent is not None:
            underline = '____________________'

            if not instances:
                instances = self.test_set
            nr_inst = len(instances)
            avg_f = 0

            def solve(incumbent, metric, engine,
                      mode, pddl_problem, conn):
                f = \
                    gaci.run_engine_config(incumbent,
                                           metric, engine,
                                           mode, pddl_problem,
                                           timelimit=planner_timelimit)

                conn.send(f)  # Send result via pipe
                conn.close()

                return f

            for inst in instances:
                if metric == 'runtime':
                    start = timeit.default_timer()

                try:
                    if isinstance(inst, tuple):
                        instance_p = f'{inst[0]}'
                        domain = f'{inst[1]}'
                    else:
                        instance_p = f'{inst}'
                        domain_path = instance_p.rsplit('/', 1)[0]
                        domain = f'{domain_path}/domain.pddl'
                    pddl_problem = self.reader.parse_problem(f'{domain}',
                                                             f'{instance_p}')

                    print('\n\nEvaluatng instance:', instance_p)

                    parent_conn, child_conn = multiprocessing.Pipe()

                    process = \
                        multiprocessing.Process(
                            target=solve, args=(incumbent, metric, engine,
                                                mode, pddl_problem,
                                                child_conn))
                    process.start()
                    child_conn.close()
                    pr_start = time.time()
                    process_pid = process.pid
                    uproc = psutil.Process(process_pid)

                    # Managing Processes and Memory, making sure system does
                    # not crash
                    jobid = os.getenv("SLURM_JOB_ID")
                    if jobid is not None:
                        try:
                            mem_res = subprocess.run(
                                ["scontrol", "show", "job", jobid],
                                capture_output=True, text=True, check=True
                            )
                            for line in mem_res.stdout.split("\n"):
                                if "mem=" in line:
                                    if 'G' in line:
                                        mult = 1000
                                    else:
                                        mult = 1
                                    mem = line.split(',')[1]
                                    mem = int(mem.split('=')[1][:-1]) * mult
                                    print('Allocated Memory is:', mem,
                                          flush=True)

                        except Exception as e:
                            print(f"Error retrieving allocated memory: {e}",
                                  flush=True)
                            mem = None
                    else:
                        mem = psutil.virtual_memory().total / (1024 * 1024)

                    while (time.time() - pr_start <= planner_timelimit + 1
                           and uproc.is_running()
                           and uproc.status() != psutil.STATUS_ZOMBIE) or (
                               time.time() - pr_start >= planner_timelimit):
                        time.sleep(1)
                        kill_times = [20, 30, 40]
                        mem_used = uproc.memory_info().rss / (1024 * 1024)
                        for child in uproc.children(recursive=True):
                            mem_used += uproc.memory_info().rss / (1024 * 1024)
                        if mem_used > 0.5 * mem:
                            for i in range(3):
                                try:
                                    self.trigger_gc_for_children(process_pid)
                                except Exception as e:
                                    print(e)
                                self.kill_process_tree(process_pid,
                                                       kill_times[i])
                            try:
                                process.kill()
                            except Exception as e:
                                print(e)
                                pass
                            continue

                    f = None

                    if parent_conn.poll(1):
                        try:
                            f = parent_conn.recv()
                        except Exception as e:
                            print(e)

                    process.join()

                except (AssertionError, NotImplementedError,
                        UPProblemDefinitionError) as err:
                    if self.verbose:
                        print('\n** Error in planning engine!')
                        print(err)
                    if metric == 'runtime':
                        f = planner_timelimit
                    elif metric == 'quality':
                        f = crash_cost

                if metric == 'runtime':
                    if f is None or f == 'unsolvable':
                        f = planner_timelimit
                    elif 'measure':
                        f = timeit.default_timer() - start
                        if f > planner_timelimit:
                            f = planner_timelimit
                if metric == 'quality' and (f is None or f == 'unsolvable'):
                    f = crash_cost

                if f is not None:
                    avg_f += f
                else:
                    if metric == 'quality':
                        avg_f += self.crash_cost
                    elif metric == 'runtime':
                        avg_f += planner_timelimit
                if metric == 'runtime':
                    if self.verbose:
                        print(f'\nFeedback on instance {inst}:\n\n', f, '\n')
                        print(underline)
                elif metric == 'quality':
                    if f is not None:
                        if self.verbose:
                            print(f'\nFeedback on instance {inst}:\n\n', f,
                                  '\n')
                            print(underline)
                    else:
                        if self.verbose:
                            print(f'\nFeedback on instance {inst}:\n\n', None,
                                  '\n')
                            print(underline)

            if nr_inst != 0:
                avg_f = avg_f / nr_inst
                if metric == 'runtime':
                    print(f'\nAverage performance on {nr_inst} instances:',
                          avg_f, 'seconds\n')
                if metric == 'quality':
                    print(f'''\nAverage quality performance on {nr_inst} 
                          instances:''',
                          avg_f, '\n')
                return avg_f
            else:
                print('\nPerformance could not be evaluated. No plans found.')
                return None
        else:
            return None

    def save_config(self, path, config, gaci, engine, plantype):
        """
        Save configuration in json file.

        :param path: Path where to save.
        :type path: str
        :param config: Configuration to save.
        :type config: dict
        :param gaci: AC interface object.
        :type gaci: ACInterface
        :param engine: Engine name.
        :type engine: str
        """
        if config is not None:
            config = gaci.transform_conf_from_ac(engine, config, plantype)
            with open(f'{path}/incumbent_{engine}.json', 'w') as f:
                json.dump(config, f)
            if self.verbose:
                print('\nSaved best configuration in ' +
                      f'{path}/incumbent_{engine}.json\n')
        else:
            if self.verbose:
                print(f'No configuration was saved. It was {config}')



from sys import platform
from urllib import request as req
import zipfile
import os
import shutil
import stat


def get_OAT():
    # Set path to up-ac
    try:
        import up_ac
        path = '/' + os.path.abspath(up_ac.__file__).strip('/__init__.py')
    except ImportError:
        path = os.getcwd().rsplit('up_ac', 1)[0]
        if path[-1] != "/":
            path += "/"
        path += 'up_ac'

    # Path to OAT directory
    save_as = f'{path}/OAT/OAT.zip'

    # Check platform
    if platform in ("linux", "linux2"):
        url = 'https://docs.optano.com/algorithm.tuner/current/' + \
            'OPTANO.Algorithm.Tuner.Application.2.1.0_linux-x64.zip'
    elif platform == "darwin":
        url = 'https://docs.optano.com/algorithm.tuner/current/' + \
            'OPTANO.Algorithm.Tuner.Application.2.1.0_osx-x64.zip'
    elif platform == "win32":
        url = 'https://docs.optano.com/algorithm.tuner/current/' + \
            'OPTANO.Algorithm.Tuner.Application.2.1.0_win-x64.zip'

    # Make OAT directory in up-ac
    if not os.path.isdir(f'{path}/OAT'):
        os.mkdir(f'{path}/OAT')

    ua_1 = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36'
    ua_2 = '(KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36'

    headers = {'User-Agent':
               f'{ua_1} {ua_2}',
               'Accept': 'application/json'}

    request = req.Request(url, headers=headers)

    # Download binaries, if not in up-ac/OAT
    if not os.path.isfile(f'{path}/OAT/OAT.zip'):
        with req.urlopen(request) as response, open(save_as, 'wb') as out_file:
            out_file.write(response.read())

    # Unzip and delete .zip
    if os.path.isfile(f'{path}/OAT/OAT.zip') and \
            not os.path.isfile(
                f'{path}/OAT/Optano.Algorithm.Tuner.Application'):

        with zipfile.ZipFile(f'{path}/OAT/OAT.zip', 'r') as zip_ref:
            zip_ref.extractall(f'{path}/OAT')
        if os.path.isfile(f'{path}/OAT/OAT.zip'):
            os.remove(f'{path}/OAT/OAT.zip')

    st = os.stat(f'{path}/OAT/Optano.Algorithm.Tuner.Application')
    os.chmod(f'{path}/OAT/Optano.Algorithm.Tuner.Application',
             st.st_mode | stat.S_IEXEC)


def copy_call_engine_OAT():
    # Copy call_engine_OAT from utils to OAT directory
    try:
        import up_ac
        path = '/' + os.path.abspath(up_ac.__file__).strip('/__init__.py')
    except ImportError:
        path = os.getcwd().rsplit('up_ac', 1)[0]
        if path[-1] != "/":
            path += "/"
        path += 'up_ac'
    
    shutil.copy(f'{path}/utils/call_engine_OAT.py', f'{path}/OAT/') 


def delete_OAT():
    # Set path to up-ac
    try:
        import up_ac
        path = '/' + os.path.abspath(up_ac.__file__).strip('/__init__.py')
    except ImportError:
        path = os.getcwd().rsplit('up_ac', 1)[0]
        if path[-1] != "/":
            path += "/"
        path += 'up_ac'

    if os.path.isdir(f'{path}/OAT/'):
        shutil.rmtree(f'{path}/OAT/')



"""Functionalities for managing and calling configurators."""

import timeit
import os
import sys
from threading import Thread
from queue import Queue
import subprocess
import dill 
import shutil
from unified_planning.exceptions import UPProblemDefinitionError, UPException

import signal
from contextlib import contextmanager
from threading import Timer

sys.path.append(os.path.abspath(os.path.dirname(__file__)))
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from AC_interface import *
from configurators import Configurator


class TimeoutException(Exception):
    pass


@contextmanager
def time_limit(seconds):
    def signal_handler(signum, frame):
        raise TimeoutException("Timed out!")
    signal.signal(signal.SIGALRM, signal_handler)
    signal.alarm(seconds)
    try:
        yield
    finally:
        signal.alarm(0)


class OATConfigurator(Configurator):
    """Configurator functions."""

    def __init__(self):
        """Initialize OAT configurator."""
        Configurator.__init__(self)

    def get_OAT_incumbent(self):

        path = self.scenario['path_to_OAT']
        read_param = False
        config = {}
        if os.path.isfile(f'{path}tunerLog.txt'):
            with open(f'{path}tunerLog.txt', 'r') as f:
                for line in f:
                    line = line.split(' ')
                    for i, l in enumerate(line):
                        line[i] = l.replace(' ', '')
                    if 'results' in line:
                        read_param = False
                    if read_param:
                        param_name = str(line[0].replace('\t', '')[:-1])
                        config[param_name] = line[1].replace('\n', '')      
                    if 'according' in line:
                        read_param = True

        else:
            config = {'No config': None}
                    
        return config

    def write_OAT_generation_status(self):

        path = self.scenario['path_to_OAT']
        config = {}
        if os.path.isfile(f'{path}tunerLog.txt'):
            with open(f'{path}tunerLog.txt', 'r') as f:
                for line in f:
                    line = line.split(' ')
                    for i, l in enumerate(line):
                        line[i] = l.replace(' ', '')
                    if 'Finished' in line:
                        if 'generation' in line:
                            print('OAT stopped at configuration_time!')
                            print('Best configuration was at generation',
                                  line[2], '/', line[4])
                        else:
                            print('OAT finished the AC run!')
        else:
            config = {'No config': None}
                    
        return config

    def get_feedback_function(self, gaci, engine, metric, mode,
                              gray_box=False):
        """
        Generate the function to run the engine and obtain feedback.

        :param gaci: AC interface object.
        :param str engine: Engine name.
        :param str metric: Optimization metric ('runtime' or 'quality').
        :param str mode: Type of planning.
        :param bool gray_box: True if gray box to use.

        :return: Planner feedback function.
        :rtype: function
        """
        if engine in self.capabilities[metric][mode]:
            self.metric = metric

            if gray_box:
                class gb_out():
                    def __init__(self, q, res):
                        self.q = q
                        self.res = res

                    def write(self, txt):
                        # TODO
                        # pass output to configurator
                        if self.res.empty():
                            self.q.put(txt)
                q = Queue()
                res = Queue()
                gb_out = gb_out(q, res)

            def planner_feedback(config, instance, reader):
                """
                Planner feedback function.

                :param config: Configuration for the planner.
                :param instance: Problem instance.
                :param reader: Reader object for parsing the problem.

                :return: Feedback value based on the planner's performance.
                :rtype: float
                """
                try:
                    import up_ac
                    path = '/' + os.path.abspath(up_ac.__file__).strip('/__init__.py')
                except ImportError:
                    path = os.getcwd().rsplit('up_ac', 1)[0]
                    if path[-1] != "/":
                        path += "/"
                    path += 'up_ac'

                sys.path.append(r"{}".format(path))

                self.reader = reader 
                
                start = timeit.default_timer()
                instance_p = f'{instance}'

                # gray box in OAT only works with runtime scenarios
                if gray_box:
                    def planner_thread(gb_out, problem, res,
                                       config, metric, engine, mode, 
                                       pddl_problem):
                        res.put(
                            gaci.run_engine_config(config,
                                                   metric,
                                                   engine,
                                                   mode,
                                                   pddl_problem,
                                                   gb_out))

                    thread = Thread(target=planner_thread,
                                    args=(gb_out, pddl_problem, res,
                                          config, metric, engine, mode, 
                                          pddl_problem),
                                    daemon=True)

                    thread.start()

                    while thread.is_alive():
                        try:
                            output = q.get(False)
                        except Exception as e:
                            print(e)
                            output = None
                        if output is not None and len(output) not in (0, 1):
                            if self.verbose:
                                print('Gray box output:', output)
                        if not res.empty():
                            thread.join()

                    feedback = res.get()

                else:
                    if metric == 'quality':
                        timelimit = self.planner_timelimit - self.patience
                    else:
                        timelimit = self.planner_timelimit
                    '''
                    feedback = \
                        gaci.run_engine_config(config,
                                               metric,
                                               engine,
                                               mode,
                                               pddl_problem)
                                               
                    try:
                        @concurrent.process(timeout=self.scenario['timelimit'])
                        def solve(config, metric, engine,
                                  mode, pddl_problem):
                            feedback = \
                                gaci.run_engine_config(config,
                                                       metric, engine,
                                                       mode, pddl_problem)

                            return feedback

                        feedback = solve(config, metric, engine,
                                         mode, pddl_problem)
                    
                        try:
                            feedback = feedback.result()
                        except TimeoutError:
                            if metric == 'runtime':
                                feedback = self.planner_timelimit
                            elif metric == 'quality':
                                feedback = self.crash_cost
                    '''
                    try:

                        if ',' in instance_p:
                            inst_tuple = instance_p.split(',')
                            domain = inst_tuple[1][2:-2]
                            instance_p = inst_tuple[0][2:-1]
                            pddl_problem = \
                                self.reader.parse_problem(f'{domain}',
                                                          f'{instance_p}')
                        else:
                            domain_path = instance_p.rsplit('/', 1)[0]
                            domain = f'{domain_path}/domain.pddl'
                            pddl_problem = self.reader.parse_problem(f'{domain}',
                                                                     f'{instance_p}')

                        def solve(config, metric, engine,
                                  mode, pddl_problem):
                            feedback = \
                                gaci.run_engine_config(config,
                                                       metric, engine,
                                                       mode, pddl_problem,
                                                       timelimit)

                            return feedback

                        try:
                            with time_limit(timelimit):
                                feedback = solve(config, metric, engine,
                                                 mode, pddl_problem)

                        except TimeoutException:  # TimeoutError:
                            if metric == 'runtime':
                                feedback = timelimit
                            elif metric == 'quality':
                                feedback = self.crash_cost

                    except (AssertionError, NotImplementedError,
                            UPProblemDefinitionError, UPException,
                            UnicodeDecodeError) as err:
                        if self.verbose:
                            print('\n** Error in planning engine!', err)
                        sys.exit('Error.')

                if feedback == 'unsolvable':
                    if metric == 'runtime':
                        sys.exit('Bad config.')
                    elif metric == 'quality':
                        feedback = self.crash_cost
                    return feedback

                elif feedback is not None:
                    if metric == 'quality':
                        self.print_feedback(engine, instance_p, feedback)
                        return feedback
                    elif metric == 'runtime':
                        if engine in ('tamer', 'pyperplan'):
                            feedback = timeit.default_timer() - start
                        self.print_feedback(engine, instance_p, feedback)
                        return feedback

                else:
                    sys.exit('Feedback is None.')

            path_to_OAT = 'path_to_OAT'

            dill.dump(
                planner_feedback, open(
                    f'{self.scenario[path_to_OAT]}feedback.pkl', 'wb'),
                recurse=True)

            planner_feedback = f'{self.scenario[path_to_OAT]}call_engine_OAT.py'

            return planner_feedback
        else:
            print(f'Algorithm Configuration for {metric} of {engine} in' + \
                  ' {mode} is not supported.')
            return None

    def set_scenario(self, engine, param_space, gaci,
                     configuration_time=120, n_trials=400, min_budget=1,
                     max_budget=3, crash_cost=0, planner_timelimit=30,
                     n_workers=1, instances=[], instance_features=None,
                     metric='runtime', popSize=128, evalLimit=2147483647,
                     tourn_size=8, winner_percent=0.125, racing=False,
                     numGens=None, goalGen=None, patience=10):
        """
        Set up algorithm configuration scenario.

        :param str engine: Engine name.
        :param param_space: ConfigSpace object.
        :param gaci: AC interface object.
        :param int configuration_time: Overall configuration time budget.
        :param int n_trials: Maximum number of engine evaluations.
        :param int min_budget: Minimum number of instances to use.
        :param int max_budget: Maximum number of instances to use.
        :param int crash_cost: Cost to use if engine fails.
        :param int planner_timelimit: Maximum runtime per evaluation.
        :param int n_workers: Number of cores to utilize.
        :param list instances: Problem instance paths.
        :param instance_features: Dictionary of instance names and lists of features.
        :param str metric: Optimization metric.
        :param int popSize: Population size of configs per generation (OAT).
        :param int evalLimit: Maximum number of evaluations (OAT).
        """
        self.crash_cost = crash_cost
        self.patience = patience
        if metric == 'quality':
            self.planner_timelimit = planner_timelimit + patience
        else:
            self.planner_timelimit = planner_timelimit

        param_file = gaci.get_ps_oat(param_space)

        try:
            import up_ac
            path = '/' + os.path.abspath(up_ac.__file__).strip('/__init__.py')
        except ImportError:
            path = os.getcwd().rsplit('up_ac', 1)[0]
            if path[-1] != "/":
                path += "/"
            path += 'up_ac'

        path_to_xml = f'{path}/OAT/{engine}parameterTree.xml'
        oat_dir = f'{path}/OAT/'

        with open(path_to_xml, 'w') as xml:
            xml.write(param_file)

        inst_dir = f'{path}/OAT/{engine}'
        if os.path.isdir(inst_dir):
            shutil.rmtree(inst_dir, ignore_errors=True)

        os.mkdir(inst_dir)
        
        if not instances:
            for ts in self.train_set:
                instances.append(str(ts))

        file_name = 0
        for inst in instances:
            with open(f'{inst_dir}/{file_name}.txt', 'w') as f:
                f.write(f'{inst}')
            file_name += 1

        if numGens is None:
            numGens = int(((configuration_time / planner_timelimit) / 2) * 0.85)
            print('Number generations to execute set to:', numGens)
            if goalGen is None:
                goalGen = int(numGens * 0.8)
                print('First generation to run on max_budget instances set to:', goalGen)
        elif goalGen is None:
            goalGen = int(numGens * 0.7)

        scenario = dict(
            xml=path_to_xml,
            timelimit=planner_timelimit,
            wallclock=configuration_time,
            start_gen=min_budget,
            end_gen=max_budget,
            n_workers=n_workers,
            metric=metric,
            instance_dir=inst_dir,
            path_to_OAT=oat_dir,
            popSize=popSize,
            evalLimit=evalLimit,
            tourn_size=tourn_size,
            winner_percent=winner_percent,
            racing=racing,
            numGens=numGens,
            goalGen=goalGen
        )

        self.scenario = scenario

    def optimize(self, feedback_function=None, gray_box=False):
        """
        Run the algorithm configuration.

        :param function feedback_function: Function to run the engine and get feedback.
        :param bool gray_box: True if gray box usage.

        :return: Tuple containing the best configuration found and None.
        :rtype: tuple
        """
        if feedback_function is not None:

            if self.verbose:
                print('\nStarting Parameter optimization\n')

            if self.scenario['metric'] == 'quality':
                tunefor = ' --byValue --ascending=true '
            elif self.scenario['metric'] == 'runtime':
                if self.scenario['racing']:
                    tunefor = ' --enableRacing=true '
                else:
                    tunefor = ' --enableRacing=false '
            
            n_workers = self.scenario['n_workers']
            path_to_OAT = self.scenario['path_to_OAT']
            param_space = self.scenario['xml']
            instance_folder = self.scenario['instance_dir']
            self.planner_timelimit = self.scenario['timelimit']
            min_budget = self.scenario['start_gen']
            max_budget = self.scenario['end_gen']
            evalLimit = self.scenario['evalLimit']
            popSize = self.scenario['popSize']
            tourn_size = self.scenario['tourn_size']
            winner_percent = self.scenario['winner_percent']
            numGens = self.scenario['numGens']
            goalGen = self.scenario['goalGen']

            try:
                p = subprocess.Popen(['./Optano.Algorithm.Tuner.Application' +
                                      ' --master' +
                                      f' --maxParallelEvaluations={n_workers} ' +
                                      '--basicCommand=\"python3 ' +
                                      f'{feedback_function} {{instance}} ' +
                                      '{{arguments}}\"' +
                                      f' --parameterTree={param_space} ' +
                                      '--trainingInstanceFolder' +
                                      f'=\"{instance_folder}\" ' +
                                      f'--cpuTimeout={self.planner_timelimit}' +
                                      f'{tunefor}' +
                                      f'--numGens={numGens} ' +
                                      f'--goalGen={goalGen} ' +
                                      f'--instanceNumbers={min_budget}:' +
                                      f'{max_budget} ' +
                                      f'--evaluationLimit={evalLimit} ' +
                                      f'--popSize={popSize} ' +
                                      f'--miniTournamentSize={tourn_size} ' +
                                      f'--winnerPercentage={winner_percent}'],
                                     stdout=subprocess.PIPE, stdin=subprocess.PIPE,
                                     cwd=f'{path_to_OAT[:-1]}', shell=True)

                timer = Timer(self.scenario['wallclock'], p.kill)

                timer.start()

                while p.poll() is None:
                    line = p.stdout.readline()
                    if self.verbose:
                        print(line.decode('utf-8'))

            finally:
                timer.cancel()

            self.write_OAT_generation_status()

            self.incumbent = self.get_OAT_incumbent()

            if self.verbose:
                print('\nBest Configuration found is:\n',
                      self.incumbent)

            return self.incumbent, None
        else:
            return None, None



"""This module includes the necessary patches."""
import pyparsing
from ConfigSpace.read_and_write.pcs import (
    ConfigurationSpace,
    UniformIntegerHyperparameter,
    UniformFloatHyperparameter,
    CategoricalHyperparameter,
    pp_cont_param,
    pp_condition,
    pp_cat_param,
    pp_forbidden_clause,
    ForbiddenEqualsClause,
    ForbiddenAndConjunction,
    OrderedDict,
    EqualsCondition,
    InCondition,
    AndConjunction
)


def patch_pcs(pcs):
    """Patch pcs.

    parameter pcs: module.
    
    return pcs (patched module)
    """
    pcs.pp_param_name = pyparsing.Word(
        pyparsing.alphanums + "_" + "-" + "@" + "." + ":" + ";" + "\\" + 
        "=" + "/" + "?" + "!" + "$" + "%" + "&" + "*" + "+" + "<" + ">" + 
        "("  "((" + ")" + "))")
    pcs.pp_choices = pcs.pp_param_name + \
        pyparsing.Optional(pyparsing.OneOrMore("," + pcs.pp_param_name))
    pcs.pp_cat_param = pcs.pp_param_name + "{" + pcs.pp_choices + "}" + \
        "[" + pcs.pp_param_name + "]"

    def _new_read(pcs_string) -> ConfigurationSpace:
        """
        Read in a 
        :py:class:`~ConfigSpace.configuration_space.ConfigurationSpace`
        definition from a pcs file.


        .. code:: python

            from ConfigSpace import ConfigurationSpace
            from ConfigSpace.read_and_write import pcs

            cs = ConfigurationSpace({"a": [1, 2, 3]})
            with open('configspace.pcs', 'w') as f:
                 f.write(pcs.write(cs))

            with open('configspace.pcs', 'r') as f:
                deserialized_conf = pcs.read(f)

        Parameters
        ----------
        pcs_string : Iterable[str]
            ConfigSpace definition in pcs format as an iterable of strings

        Returns
        -------
        :py:class:`~ConfigSpace.configuration_space.ConfigurationSpace`
            The deserialized ConfigurationSpace object
        """
        if isinstance(pcs_string, str):
            pcs_string = pcs_string.split("\n")

        configuration_space = ConfigurationSpace()
        conditions = []
        forbidden = []

        # some statistics
        ct = 0
        cont_ct = 0
        cat_ct = 0
        line_ct = 0

        for line in pcs_string:
            line_ct += 1

            if "#" in line:
                # It contains a comment
                pos = line.find("#")
                line = line[:pos]

            # Remove quotes and whitespaces at beginning and end
            line = line.replace('"', "").replace("'", "")
            line = line.strip()

            if "|" in line:
                # It's a condition
                try:
                    c = pp_condition.parseString(line)
                    conditions.append(c)
                except pyparsing.ParseException as e:
                    raise NotImplementedError(
                        f"Could not parse condition: {line}") from e

                continue
            if "}" not in line and "]" not in line:
                continue
            if line.startswith("{") and line.endswith("}"):
                forbidden.append(line)
                continue
            if len(line.strip()) == 0:
                continue

            ct += 1
            param = None

            create = {
                "int": UniformIntegerHyperparameter,
                "float": UniformFloatHyperparameter,
                "categorical": CategoricalHyperparameter,
            }

            try:
                param_list = pp_cont_param.parseString(line)
                il = param_list[9:]
                if len(il) > 0:
                    il = il[0]
                param_list = param_list[:9]
                name = param_list[0]
                lower = float(param_list[2])  # type: ignore
                upper = float(param_list[4])  # type: ignore
                paramtype = "int" if "i" in il else "float"
                log = "l" in il
                default_value = float(param_list[7])  # type: ignore
                param = create[paramtype](
                    name=name,
                    lower=lower,
                    upper=upper,
                    q=None,
                    log=log,
                    default_value=default_value,
                )
                cont_ct += 1
            except pyparsing.ParseException:
                pass

            try:
                param_list = pp_cat_param.parseString(line)
                name = param_list[0]
                choices = list(param_list[2:-4:2])
                default_value = param_list[-2]
                param = create["categorical"](
                    name=name, choices=choices, default_value=default_value)
                cat_ct += 1
            except pyparsing.ParseException:
                pass

            if param is None:
                raise NotImplementedError("Could not parse: %s" % line)

            configuration_space.add_hyperparameter(param)

        for clause in forbidden:
            # TODO test this properly!
            # TODO Add a try/catch here!
            # noinspection PyUnusedLocal
            param_list = pp_forbidden_clause.parseString(clause)
            tmp_list: list = []
            clause_list = []
            for value in param_list[1:]:
                if len(tmp_list) < 3:
                    tmp_list.append(value)
                else:
                    # So far, only equals is supported by SMAC
                    if tmp_list[1] == "=":
                        # TODO maybe add a check if the hyperparameter is
                        # actually in the configuration space
                        if isinstance(configuration_space[tmp_list[0]],
                                      UniformIntegerHyperparameter):
                            vc_val = int(tmp_list[2])
                        elif isinstance(configuration_space[tmp_list[0]],
                                        UniformFloatHyperparameter):
                            vc_val = float(tmp_list[2])
                        else:
                            vc_val = tmp_list[2]
                        clause_list.append(
                            ForbiddenEqualsClause(
                                configuration_space[tmp_list[0]], vc_val),
                        )
                    else:
                        raise NotImplementedError()
                    tmp_list = []
            configuration_space.add_forbidden_clause(
                ForbiddenAndConjunction(*clause_list))

        # Now handle conditions
        # If there are two conditions for one child, these two conditions 
        # are an AND-conjunction of conditions, thus we have to connect them
        conditions_per_child: dict = OrderedDict()
        for condition in conditions:
            child_name = condition[0]
            if child_name not in conditions_per_child:
                conditions_per_child[child_name] = []
            conditions_per_child[child_name].append(condition)

        for child_name in conditions_per_child:
            condition_objects = []
            for condition in conditions_per_child[child_name]:
                child = configuration_space[child_name]
                parent_name = condition[2]
                parent = configuration_space[parent_name]
                restrictions = condition[5:-1:2]

                # TODO: cast the type of the restriction!
                if len(restrictions) == 1:
                    if isinstance(parent, UniformIntegerHyperparameter):
                        restrictions[0] = int(restrictions[0])
                    elif isinstance(parent, UniformFloatHyperparameter):
                        restrictions[0] = float(restrictions[0])
                    condition = EqualsCondition(child, parent, restrictions[0])
                else:
                    if isinstance(parent, UniformIntegerHyperparameter):
                        for i, r in enumerate(restrictions):
                            restrictions[i] = int(r)
                    elif isinstance(parent, UniformFloatHyperparameter):
                        for i, r in enumerate(restrictions):
                            restrictions[i] = float(r)
                    condition = InCondition(child, parent, values=restrictions)
                condition_objects.append(condition)

            # Now we have all condition objects for this child,
            # so we can build a giant AND-conjunction of them
            # (if number of conditions >= 2)!

            if len(condition_objects) > 1:
                and_conjunction = AndConjunction(*condition_objects)
                configuration_space.add_condition(and_conjunction)
            else:
                configuration_space.add_condition(condition_objects[0])

        return configuration_space

    pcs.read = _new_read

    return pcs



"""Monkey patches engine classes to be parameterizable where needed."""


def parametrizable_engines(engine):
    if engine == 'lpg-anytime':
        from up_lpg.lpg_planner import LPGAnytimeEngine
        original_init = LPGAnytimeEngine.__init__

        def new_lpg_anytime_init(self, **kwargs):
            original_init(self)
            self.parameter = []
            for param, val in kwargs.items():
                self.parameter.append(str(param))
                if val != '':
                    self.parameter.append(str(val))

        LPGAnytimeEngine.__init__ = new_lpg_anytime_init



"""Functions to get feedback from engines performance."""


def qaul_feedback(engine, result):
    """Transform/parse specific solution quality engine output.

    parameter engine: str, name of engine.
    parameter result: object, planning result.
    """
    feedback = None

    if engine in ('lpg', 'lpg-anytime'):
        feedback = \
            float(
                next((m for m in result.log_messages[0].message.split('\n')
                     if 'Plan quality' in m), None).split(' ')[-2])

    elif engine in ('fast-downward', 'symk'):
        feedback = \
            float(
                next((m for m in result.log_messages[0].message.split('\n')
                     if 'Plan cost' in m), None).split(' ')[5])

    elif engine in ('enhsp', 'enhsp-any'):
        feedback = \
            float(
                next((m for m in result.log_messages[0].message.split('\n')
                     if 'Metric' in m), None).split(':')[1])

    return feedback


def runtime_feedback(engine, result):
    """Transform/parse specific solution runtime engine output.

    parameter engine: str, name of engine.
    parameter result: object, planning result.
    """
    feedback = None

    if engine == 'lpg':
        output = result.log_messages[0].message.split('\n')
        for line in output:
            if 'Total time:' in line:
                feedback = float(line.split(' ')[-1])

    elif engine in ('fast-downward', 'symk'):
        output = result.log_messages[0].message.split('\n')
        for line in output:
            if 'Planner time:' in line:
                feedback = float(line.split(' ')[-1][:-1])

    elif engine == 'enhsp':
        output = result.log_messages[0].message.split('\n')
        for line in output:
            if 'Planning Time' in line:
                feedback = (float(line.split(':')[1])) / 1000  # msec

    elif engine in ('pyperplan', 'tamer'):
        feedback = 'measure'

    return feedback


def gray_box_feedback(engine, result):
    """Transform/parse intermediate output.

    parameter engine: str, name of engine.
    parameter result: object, planning result.
    """
    # TODO
    if engine in ('lpg', 'lpg-anytime'):
        feedback = None

    elif engine in ('fast-downward', 'symk'):
        feedback = None

    return feedback



#!/usr/bin/env python

import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(name='up_ac',
      version='0.0.1.1',
      long_description=README,
      long_description_content_type="text/markdown",
      description='An Algorithm Configuration package for unified-planning.',
      url='https://github.com/DimitriWeiss/up-ac.git',
      author='Dimitri Weiß',
      author_email='dimitri-weiss@web.com',
      packages=find_packages(exclude=["*.tests"]),
      package_data={'': ['test_problems/citycar/*',
                         'test_problems/counters/*',
                         'test_problems/depot/*',
                         'test_problems/htn-transport/*',
                         'test_problems/matchcellar/*',
                         'test_problems/miconic/*',
                         'test_problems/robot_fastener/*',
                         'test_problems/safe_road/*',
                         'test_problems/sailing/*',
                         'test_problems/visit_precedence/*',
                         'engine_pcs/*']},
      include_package_data=True,
      install_requires=["unified-planning", "ConfigSpace",
                        "tarski", "pebble", "dill"],
      license='LICENSE.txt',
      )



from unified_planning.exceptions import UPProblemDefinitionError, UPException
import argparse
import ast
import os
import sys
from unified_planning.io import PDDLReader
import unified_planning as up
import timeit
import signal
from contextlib import contextmanager


reader = PDDLReader()
up.shortcuts.get_environment().credits_stream = None


class TimeoutException(Exception):
    pass


@contextmanager
def time_limit(seconds):
    def signal_handler(signum, frame):
        raise TimeoutException("Timed out!")
    signal.signal(signal.SIGALRM, signal_handler)
    signal.alarm(seconds)
    try:
        yield
    finally:
        signal.alarm(0)


def sel_call_up(instance, config):

    try:
        import up_ac
        path = '/' + os.path.abspath(up_ac.__file__).strip('/__init__.py')
        up_ac_path = path
        sys.path.append(r"{}".format(path))
        sys.path.append(r"{}".format(path + '/utils'))
        path += '/utils'
    except ImportError:
        path = os.getcwd().rsplit('up_ac', 1)[0]
        if path[-1] != "/":
            path += "/"
        up_ac_path = path + 'up_ac'
        sys.path.append(r"{}".format(up_ac_path))
        sys.path.append(r"{}".format(path + 'up_ac/utils'))
        path += 'up_ac/utils'

    setting_path = f'{path}/sel_feedback_setting'
    reader._env.error_used_name = False

    try:
        from up_ac.Selector_interface import SelectorInterface
    except (ModuleNotFoundError, ImportError, Exception):
        from Selector_interface import SelectorInterface

    with open(f'{setting_path}/feedback_args.txt', 'r') as f:
        feedback_args = ast.literal_eval(f.read())

    selgaci = SelectorInterface()
    selgaci.read_engine_pcs(
        [feedback_args['engine']], f'{up_ac_path}/engine_pcs')

    feedback = planner_feedback(config, instance, reader,
                                selgaci, feedback_args)

    if feedback_args['metric'] == 'runtime':
        if feedback != feedback_args['timelimit'] and feedback is not None:
            print('Tuned plan runtime is:', feedback)
        else:
            print('Fail! Feedback is:', feedback)
    elif feedback_args['metric'] == 'quality':
        print('Tuned plan quality is:', feedback)


def print_feedback(engine, instance, feedback, verbose):
    if verbose:
        print(f'** Feedback of {engine} on instance\n**' +
              f' {instance}\n** is {feedback}\n\n')


def planner_feedback(config, instance, reader, gaci, fa):
    start = timeit.default_timer()
    instance_p = f'{instance}'

    engine = fa['engine']
    metric = fa['metric']
    mode = fa['mode']
    gray_box = fa['gray_box']
    patience = fa['patience']
    train_set = fa['train_set']
    planner_timelimit = fa['timelimit']
    crash_cost = fa['crash_cost']
    verbose = fa['verbose']

    if metric == 'quality':
        timelimit = planner_timelimit - patience
    else:
        timelimit = planner_timelimit

    try:
        if isinstance(train_set, dict) and \
                isinstance(train_set[instance_p], tuple):
            domain = train_set[instance_p][1]
            problem = train_set[instance_p][0]
            pddl_problem = reader.parse_problem(domain, problem)
        else:
            domain_path = instance_p.rsplit('/', 1)[0]
            domain = f'{domain_path}/domain.pddl'
            pddl_problem = reader.parse_problem(f'{domain}',
                                                f'{instance_p}')
        '''
        if engine == 'tamer' or engine == 'pyperplan':
            @concurrent.process(timeout=timelimit)
            def solve(config, metric, engine,
                      mode, pddl_problem):
                feedback = \
                    gaci.run_engine_config(config,
                                           metric, engine,
                                           mode, pddl_problem,
                                           timelimit)

                return feedback
        else:
        '''
        def solve(config, metric, engine,
                  mode, pddl_problem):
            feedback = \
                gaci.run_engine_config(config,
                                       metric, engine,
                                       mode, pddl_problem,
                                       timelimit)

            return feedback

        try:
            if engine == 'tamer' or engine == 'pyperplan':
                def handler(signum, frame):
                    raise TimeoutError("Function timed out!")

                signal.signal(signal.SIGALRM, handler)
                signal.alarm(timelimit)

                try:
                    feedback = solve(config, metric, engine,
                                     mode, pddl_problem)
                except TimeoutError as te:
                    print(te)
                    feedback = timelimit
                finally:
                    signal.alarm(0)

            else:
                with time_limit(timelimit):
                    feedback = solve(config, metric, engine,
                                     mode, pddl_problem)

        except TimeoutException:
            if metric == 'runtime':
                feedback = timelimit
            elif metric == 'quality':
                feedback = crash_cost

    except (AssertionError, NotImplementedError,
            UPProblemDefinitionError, UPException,
            UnicodeDecodeError) as err:
        if verbose:
            print('\n** Error in planning engine!', err)
        if metric == 'runtime':
            feedback = timelimit
        elif metric == 'quality':
            feedback = crash_cost

    if feedback == 'unsolvable':
        if metric == 'runtime':
            feedback = timelimit
        elif metric == 'quality':
            feedback = crash_cost
  
    if feedback is not None:
        # SMAC always minimizes
        if metric == 'quality':
            print_feedback(engine, instance, feedback, verbose)
            return feedback
        # Solving runtime optimization by passing
        # runtime as result, since smac minimizes it
        elif metric == 'runtime':
            if engine in ('tamer', 'pyperplan'):
                feedback = timeit.default_timer() - start
                if feedback > timelimit:
                    feedback = timelimit
                print_feedback(engine, instance, feedback, verbose)
            else:
                if feedback > timelimit:
                    feedback = timelimit
                print_feedback(engine, instance, feedback, verbose)
            return feedback
    else:
        # Penalizing failed runs
        if metric == 'runtime':
            # Penalty is max runtime in runtime scenario
            feedback = timelimit
            print_feedback(engine, instance, feedback, verbose)
        else:
            # Penalty is defined by user in quality scenario
            feedback = crash_cost
            print_feedback(engine, instance, feedback, verbose)

        return feedback


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument('--i', type=str)
    parser.add_argument('--c', type=str)

    args = vars(parser.parse_args())

    sel_call_up(args['i'], ast.literal_eval(args['c']))



import sys
import os

sys.path.append(os.path.abspath(os.path.dirname(__file__)))
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "..")))

from Smac_interface import SmacInterface


class SelectorInterface(SmacInterface):
    """Using Smac interface."""
